#!/usr/bin/env python3
# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance Chain CLI
"""
import asyncio
import logging
from pprint import pprint

import click
from .httpclient import HTTPClient

log = logging.getLogger(__name__)


def run(coro):
    """ Run an async coroutine """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(coro)


def dex_run(method, **kwargs):
    """ Run a method on the Binance DEX client """
    dex = HTTPClient(testnet=False)
    result = run(getattr(dex, method)(**kwargs))
    pprint(result)
    run(dex.close())
    return result


@click.group()
def main():
    pass


@main.command()
def time():
    dex_run("get_time")


@main.command()
def node_info():
    dex_run("get_node_info")


@main.command()
def fees():
    dex_run("get_fees")


@main.command()
def validators():
    dex_run("get_validators")


@main.command()
def peers():
    dex_run("get_peers")


@main.command()
def tokens():
    dex_run("get_token_list")


@main.command()
def markets():
    dex_run("get_markets")


@main.command()
@click.argument("symbol")
def depth(**kwargs):
    dex_run("get_depth", **kwargs)


@main.command()
@click.argument("hash")
def broadcast(**kwargs):
    dex_run("broadcast", **kwargs)


@main.command()
@click.argument("symbol")
@click.argument("interval")
def klines(**kwargs):
    dex_run("get_klines", **kwargs)


@main.command()
@click.option("--address", help="the seller/buyer address", type=str)
@click.option("--end", help="end time in milliseconds", type=int)
@click.option("--limit", help="default 50; max 1000.", type=int)
@click.option("--offset", help="start with 0; default 0.", type=int)
@click.option("--side", help="order side. 1 for buy and 2 for sell.", type=int)
@click.option("--start", help="start time in milliseconds", type=int)
@click.option(
    "--status",
    help="order status list. Allowed value: [Ack, PartialFill, IocNoFill, FullyFill, Canceled, Expired, FailedBlocking, FailedMatching]",
    type=str,
)
@click.option("--symbol", help="symbol", type=str)
@click.option(
    "--total",
    help="total number required, 0 for not required and 1 for required; default not required, return total=-1 in response",
    type=int,
)
def closed_orders(**kwargs):
    dex_run("get_closed_orders", **kwargs)


@main.command()
@click.option("--address", help="the seller/buyer address", type=str)
@click.option("--limit", help="default 50; max 1000.", type=int)
@click.option("--offset", help="start with 0; default 0.", type=int)
@click.option("--symbol", help="symbol", type=str)
@click.option(
    "--total",
    help="total number required, 0 for not required and 1 for required; default not required, return total=-1 in response",
    type=int,
)
def open_orders(**kwargs):
    dex_run("get_open_orders", **kwargs)


@main.command()
@click.argument("symbol", default="")
def ticker(**kwargs):
    dex_run("get_ticker", **kwargs)


@main.command()
@click.option("--address", help="the seller/buyer address")
@click.option("--end", help="end time", type=int)
@click.option("--limit", help="default 50; max 1000.", type=int)
@click.option("--offset", help="start with 0; default 0.", type=int)
@click.option("--start", help="start time", type=int)
@click.option(
    "--total",
    help="total number required, 0 for not required and 1 for required; default not required, return total=-1 in response",
    type=int,
)
def trades(**kwargs):
    dex_run("get_trades", **kwargs)


@main.command()
@click.option("--address", help="the seller/buyer address")
@click.option("--end", help="end time", type=int)
@click.option("--limit", help="default 50; max 1000.", type=int)
@click.option("--offset", help="start with 0; default 0.", type=int)
@click.option("--start", help="start time", type=int)
@click.option(
    "--total",
    help="total number required, 0 for not required and 1 for required; default not required, return total=-1 in response",
    type=int,
)
def block_exchange_fee(**kwargs):
    dex_run("get_block_exchange_fee", **kwargs)


@main.command()
@click.argument("address")
@click.option("--height", help="block height", type=int)
@click.option("--end", help="end in milliseconds", type=int)
@click.option("--limit", help="default 50; max 1000.", type=int)
@click.option("--offset", help="start with 0; default 0.", type=int)
@click.option(
    "--side", help="transaction side. Allowed value: [ RECEIVE, SEND]", type=str
)
@click.option("--start", help="start time in milliseconds", type=int)
@click.option("--tx-asset", help="txAsset", type=str)
@click.option(
    "--tx-type",
    help="transaction type. Allowed value: [NEW_ORDER,ISSUE_TOKEN,BURN_TOKEN,LIST_TOKEN,CANCEL_ORDER,FREEZE_TOKEN,UN_FREEZE_TOKEN,TRANSFER,PROPOSAL,VOTE,MINT,DEPOSIT]",
    type=str,
)
def transactions(**kwargs):
    dex_run("get_transactions", **kwargs)


if __name__ == "__main__":
    main()



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance DEX SDK Test Suite for Wallet builder
"""
from pprint import pprint

import pytest
import asyncio
import json
from decimal import Decimal
from binancechain import Wallet, HTTPClient

MNEMONIC_2 = "tennis utility midnight pattern that foot security tent punch glance still night virus loop trade velvet rent glare ramp cushion defy grass section cage"
MNEMONIC = "apart conduct congress bless remember picnic aerobic nothing dinner guilt catch brain sunny vocal advice castle horror shift reject valley evoke fork syrup code"
PASSWORD = "Alkseui12p,d"
KEYSTORE = {
    "address": "309ac34e39715899735bfaed1677ff82b5bf35e0",
    "crypto": {
        "cipher": "aes-128-ctr",
        "cipherparams": {"iv": "bab8a3ace08205c897a8a69ddb9bf056"},
        "ciphertext": "6b273686d7858feac0708d64acdd9bdb353191d4f735e8c04a11ee3a31f08f0f",
        "kdf": "pbkdf2",
        "kdfparams": {
            "c": 1000000,
            "dklen": 32,
            "prf": "hmac-sha256",
            "salt": "ed2e6c05ec3e5e49c3519a07d0572cf0",
        },
        "mac": "f11eb5068e3e697c1eeeccd4c5f167c6fc0956f40b3f0cb1588ccd4f0b9d7a13",
    },
    "id": "a0c93b32-6a63-4346-aafa-afdc2b5637dd",
    "version": 3,
}

PRIVATEKEY = "14ffce9d61a219f3b24c0a093a24837368e2317b1e96d77f80505df49c2224c3"
ADDRESS = "bnb1r5jc35v338tlphnjx65wy7tecm6vm82t87qjt0"
TRANSACTION_MSG = b'{"account_number":"668107","chain_id":"Binance-Chain-Nile","data":null,"memo":"","msgs":[{"inputs":[{"address":"tbnb1r5jc35v338tlphnjx65wy7tecm6vm82tftfkt7","coins":[{"amount":10000000,"denom":"BNB"}]}],"outputs":[{"address":"tbnb1nhvpuq0u5pgpry0x2ap2hqv9n5jfkj90eps6qx","coins":[{"amount":10000000,"denom":"BNB"}]}]}],"sequence":"35","source":"1"}'


@pytest.fixture
async def chain():
    chain = HTTPClient(testnet=True)
    yield chain
    await chain.close()


@pytest.fixture
async def wallet():
    wallet = Wallet.wallet_from_privatekey(privatekey=PRIVATEKEY, password="")
    yield wallet


@pytest.mark.asyncio
async def test_create_wallet(chain):
    wallet = Wallet.create_wallet()
    address = wallet.get_address()
    print(wallet)
    assert address, "Failed to get address from wallet"


@pytest.mark.asyncio
async def test_create_wallet_mnemonic(chain):
    wallet = Wallet.create_wallet_mnemonic(password=PASSWORD)
    address = wallet.get_address()
    assert address, "Failed to get address from wallet"


@pytest.mark.asyncio
async def test_create_keystore(chain):
    keystore = Wallet.create_keystore(password=PASSWORD)
    assert keystore, "Failed to create keystore"


@pytest.mark.asyncio
async def test_wallet_from_keystore(chain):
    keystore = Wallet.create_keystore(password=PASSWORD)
    assert keystore, "Failed to create keystore"
    wallet = Wallet.wallet_from_keystore(keystore=keystore, password=PASSWORD)
    address = wallet.get_address()
    assert address, "Failed to get address from recovered wallet"


@pytest.mark.asyncio
async def test_not_recover_wallet_keystore_with_wrong_password(chain):
    keystore = Wallet.create_keystore(password=PASSWORD)
    assert keystore, "Failed to create keystore"
    try:
        wallet = Wallet.wallet_from_keystore(keystore=keystore, password="")
        assert False
    except ValueError as e:
        print(e)
        assert True


@pytest.mark.asyncio
async def test_wallet_from_mnemonic(chain):
    wallet = Wallet.wallet_from_mnemonic(MNEMONIC, password="")
    address = wallet.get_address()
    privatekey = wallet.get_privatekey()
    assert address, "Failed to get address from wallet"
    assert privatekey, "Failed to get private key from wallet"
    assert privatekey == PRIVATEKEY, "Failed to get private key from wallet"
    assert address == ADDRESS, "Not getting the same address from recover"


@pytest.mark.asyncio
async def test_wallet_from_privatekey(chain):
    wallet = Wallet.wallet_from_privatekey(privatekey=PRIVATEKEY, password="")
    address = wallet.get_address()
    assert address, "Failed to get address from wallet"
    assert address == ADDRESS, "Not getting the same address from recover"


@pytest.mark.asyncio
async def test_signing_msg(wallet, chain):
    pub, sig = wallet.sign(TRANSACTION_MSG)
    assert pub, "No public key return"
    assert sig, "No signature return"


@pytest.mark.asyncio
async def test_verify_signature(chain, wallet):
    pub, sig = wallet.sign(TRANSACTION_MSG)
    valid = wallet.verify_signature(TRANSACTION_MSG, signature=sig)
    assert pub, "No public key return"
    assert sig, "No signature return"
    assert valid, "Invalid signature"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: transaction.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='transaction.proto',
  package='transaction',
  syntax='proto3',
  serialized_pb=_b('\n\x11transaction.proto\x12\x0btransaction\"U\n\x05StdTx\x12\x0c\n\x04msgs\x18\x01 \x03(\x0c\x12\x12\n\nsignatures\x18\x02 \x03(\x0c\x12\x0c\n\x04memo\x18\x03 \x01(\t\x12\x0e\n\x06source\x18\x04 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\"\\\n\x0cStdSignature\x12\x0f\n\x07pub_key\x18\x01 \x01(\x0c\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x12\x16\n\x0e\x61\x63\x63ount_number\x18\x03 \x01(\x03\x12\x10\n\x08sequence\x18\x04 \x01(\x03\"\x05\n\x03Msg\"\x8d\x01\n\x08NewOrder\x12\x0e\n\x06sender\x18\x01 \x01(\x0c\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0e\n\x06symbol\x18\x03 \x01(\t\x12\x11\n\tordertype\x18\x04 \x01(\x03\x12\x0c\n\x04side\x18\x05 \x01(\x03\x12\r\n\x05price\x18\x06 \x01(\x03\x12\x10\n\x08quantity\x18\x07 \x01(\x03\x12\x13\n\x0btimeinforce\x18\x08 \x01(\x03\"<\n\x0b\x43\x61ncelOrder\x12\x0e\n\x06sender\x18\x01 \x01(\x0c\x12\x0e\n\x06symbol\x18\x02 \x01(\t\x12\r\n\x05refid\x18\x03 \x01(\t\"P\n\x04Send\x12\"\n\x06inputs\x18\x01 \x03(\x0b\x32\x12.transaction.Input\x12$\n\x07outputs\x18\x02 \x03(\x0b\x32\x13.transaction.Output\"&\n\x05Token\x12\r\n\x05\x64\x65nom\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x03\";\n\x05Input\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12!\n\x05\x63oins\x18\x02 \x03(\x0b\x32\x12.transaction.Token\"<\n\x06Output\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12!\n\x05\x63oins\x18\x02 \x03(\x0b\x32\x12.transaction.Token\"6\n\x06\x46reeze\x12\x0c\n\x04\x66rom\x18\x01 \x01(\x0c\x12\x0e\n\x06symbol\x18\x02 \x01(\t\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x03\"8\n\x08Unfreeze\x12\x0c\n\x04\x66rom\x18\x01 \x01(\x0c\x12\x0e\n\x06symbol\x18\x02 \x01(\t\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x03\"S\n\x04Vote\x12\x13\n\x0bproposal_id\x18\x01 \x01(\x03\x12\r\n\x05voter\x18\x02 \x01(\x0c\x12\'\n\x06option\x18\x03 \x01(\x0e\x32\x17.transaction.VoteOption*G\n\nVoteOption\x12\x0b\n\x07Unknown\x10\x00\x12\x07\n\x03Yes\x10\x01\x12\x0b\n\x07\x41\x62stain\x10\x02\x12\x06\n\x02No\x10\x03\x12\x0e\n\nNoWithVeto\x10\x04\x62\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_VOTEOPTION = _descriptor.EnumDescriptor(
  name='VoteOption',
  full_name='transaction.VoteOption',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Unknown', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Yes', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='Abstain', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='No', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NoWithVeto', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=872,
  serialized_end=943,
)
_sym_db.RegisterEnumDescriptor(_VOTEOPTION)

VoteOption = enum_type_wrapper.EnumTypeWrapper(_VOTEOPTION)
Unknown = 0
Yes = 1
Abstain = 2
No = 3
NoWithVeto = 4



_STDTX = _descriptor.Descriptor(
  name='StdTx',
  full_name='transaction.StdTx',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='msgs', full_name='transaction.StdTx.msgs', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='transaction.StdTx.signatures', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='memo', full_name='transaction.StdTx.memo', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='source', full_name='transaction.StdTx.source', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data', full_name='transaction.StdTx.data', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=119,
)


_STDSIGNATURE = _descriptor.Descriptor(
  name='StdSignature',
  full_name='transaction.StdSignature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pub_key', full_name='transaction.StdSignature.pub_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='signature', full_name='transaction.StdSignature.signature', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='account_number', full_name='transaction.StdSignature.account_number', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sequence', full_name='transaction.StdSignature.sequence', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=121,
  serialized_end=213,
)


_MSG = _descriptor.Descriptor(
  name='Msg',
  full_name='transaction.Msg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=215,
  serialized_end=220,
)


_NEWORDER = _descriptor.Descriptor(
  name='NewOrder',
  full_name='transaction.NewOrder',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sender', full_name='transaction.NewOrder.sender', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='id', full_name='transaction.NewOrder.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='transaction.NewOrder.symbol', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ordertype', full_name='transaction.NewOrder.ordertype', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='side', full_name='transaction.NewOrder.side', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='price', full_name='transaction.NewOrder.price', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='quantity', full_name='transaction.NewOrder.quantity', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timeinforce', full_name='transaction.NewOrder.timeinforce', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=223,
  serialized_end=364,
)


_CANCELORDER = _descriptor.Descriptor(
  name='CancelOrder',
  full_name='transaction.CancelOrder',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sender', full_name='transaction.CancelOrder.sender', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='transaction.CancelOrder.symbol', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='refid', full_name='transaction.CancelOrder.refid', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=366,
  serialized_end=426,
)


_SEND = _descriptor.Descriptor(
  name='Send',
  full_name='transaction.Send',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputs', full_name='transaction.Send.inputs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='outputs', full_name='transaction.Send.outputs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=428,
  serialized_end=508,
)


_TOKEN = _descriptor.Descriptor(
  name='Token',
  full_name='transaction.Token',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='denom', full_name='transaction.Token.denom', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='amount', full_name='transaction.Token.amount', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=510,
  serialized_end=548,
)


_INPUT = _descriptor.Descriptor(
  name='Input',
  full_name='transaction.Input',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='transaction.Input.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coins', full_name='transaction.Input.coins', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=550,
  serialized_end=609,
)


_OUTPUT = _descriptor.Descriptor(
  name='Output',
  full_name='transaction.Output',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='transaction.Output.address', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='coins', full_name='transaction.Output.coins', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=611,
  serialized_end=671,
)


_FREEZE = _descriptor.Descriptor(
  name='Freeze',
  full_name='transaction.Freeze',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='from', full_name='transaction.Freeze.from', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='transaction.Freeze.symbol', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='amount', full_name='transaction.Freeze.amount', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=673,
  serialized_end=727,
)


_UNFREEZE = _descriptor.Descriptor(
  name='Unfreeze',
  full_name='transaction.Unfreeze',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='from', full_name='transaction.Unfreeze.from', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='transaction.Unfreeze.symbol', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='amount', full_name='transaction.Unfreeze.amount', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=729,
  serialized_end=785,
)


_VOTE = _descriptor.Descriptor(
  name='Vote',
  full_name='transaction.Vote',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='proposal_id', full_name='transaction.Vote.proposal_id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='voter', full_name='transaction.Vote.voter', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='option', full_name='transaction.Vote.option', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=787,
  serialized_end=870,
)

_SEND.fields_by_name['inputs'].message_type = _INPUT
_SEND.fields_by_name['outputs'].message_type = _OUTPUT
_INPUT.fields_by_name['coins'].message_type = _TOKEN
_OUTPUT.fields_by_name['coins'].message_type = _TOKEN
_VOTE.fields_by_name['option'].enum_type = _VOTEOPTION
DESCRIPTOR.message_types_by_name['StdTx'] = _STDTX
DESCRIPTOR.message_types_by_name['StdSignature'] = _STDSIGNATURE
DESCRIPTOR.message_types_by_name['Msg'] = _MSG
DESCRIPTOR.message_types_by_name['NewOrder'] = _NEWORDER
DESCRIPTOR.message_types_by_name['CancelOrder'] = _CANCELORDER
DESCRIPTOR.message_types_by_name['Send'] = _SEND
DESCRIPTOR.message_types_by_name['Token'] = _TOKEN
DESCRIPTOR.message_types_by_name['Input'] = _INPUT
DESCRIPTOR.message_types_by_name['Output'] = _OUTPUT
DESCRIPTOR.message_types_by_name['Freeze'] = _FREEZE
DESCRIPTOR.message_types_by_name['Unfreeze'] = _UNFREEZE
DESCRIPTOR.message_types_by_name['Vote'] = _VOTE
DESCRIPTOR.enum_types_by_name['VoteOption'] = _VOTEOPTION

StdTx = _reflection.GeneratedProtocolMessageType('StdTx', (_message.Message,), dict(
  DESCRIPTOR = _STDTX,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.StdTx)
  ))
_sym_db.RegisterMessage(StdTx)

StdSignature = _reflection.GeneratedProtocolMessageType('StdSignature', (_message.Message,), dict(
  DESCRIPTOR = _STDSIGNATURE,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.StdSignature)
  ))
_sym_db.RegisterMessage(StdSignature)

Msg = _reflection.GeneratedProtocolMessageType('Msg', (_message.Message,), dict(
  DESCRIPTOR = _MSG,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Msg)
  ))
_sym_db.RegisterMessage(Msg)

NewOrder = _reflection.GeneratedProtocolMessageType('NewOrder', (_message.Message,), dict(
  DESCRIPTOR = _NEWORDER,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.NewOrder)
  ))
_sym_db.RegisterMessage(NewOrder)

CancelOrder = _reflection.GeneratedProtocolMessageType('CancelOrder', (_message.Message,), dict(
  DESCRIPTOR = _CANCELORDER,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.CancelOrder)
  ))
_sym_db.RegisterMessage(CancelOrder)

Send = _reflection.GeneratedProtocolMessageType('Send', (_message.Message,), dict(
  DESCRIPTOR = _SEND,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Send)
  ))
_sym_db.RegisterMessage(Send)

Token = _reflection.GeneratedProtocolMessageType('Token', (_message.Message,), dict(
  DESCRIPTOR = _TOKEN,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Token)
  ))
_sym_db.RegisterMessage(Token)

Input = _reflection.GeneratedProtocolMessageType('Input', (_message.Message,), dict(
  DESCRIPTOR = _INPUT,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Input)
  ))
_sym_db.RegisterMessage(Input)

Output = _reflection.GeneratedProtocolMessageType('Output', (_message.Message,), dict(
  DESCRIPTOR = _OUTPUT,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Output)
  ))
_sym_db.RegisterMessage(Output)

Freeze = _reflection.GeneratedProtocolMessageType('Freeze', (_message.Message,), dict(
  DESCRIPTOR = _FREEZE,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Freeze)
  ))
_sym_db.RegisterMessage(Freeze)

Unfreeze = _reflection.GeneratedProtocolMessageType('Unfreeze', (_message.Message,), dict(
  DESCRIPTOR = _UNFREEZE,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Unfreeze)
  ))
_sym_db.RegisterMessage(Unfreeze)

Vote = _reflection.GeneratedProtocolMessageType('Vote', (_message.Message,), dict(
  DESCRIPTOR = _VOTE,
  __module__ = 'transaction_pb2'
  # @@protoc_insertion_point(class_scope:transaction.Vote)
  ))
_sym_db.RegisterMessage(Vote)


# @@protoc_insertion_point(module_scope)



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance DEX SDK Test Suite
"""
from pprint import pprint
from datetime import datetime, timedelta

import aiohttp
import pytest

from binancechain import HTTPClient, BinanceChainException


@pytest.fixture
async def client():
    client = HTTPClient(testnet=True)
    yield client
    await client.close()


@pytest.mark.asyncio
async def test_time(client):
    time = await client.get_time()
    pprint(time)
    for key in ("ap_time", "block_time"):
        assert key in time


@pytest.mark.asyncio
async def test_node_info(client):
    info = await client.get_node_info()
    pprint(info)
    for key in ("node_info", "sync_info", "validator_info"):
        assert key in info


@pytest.mark.asyncio
async def test_get_validators(client):
    validators = await client.get_validators()
    pprint(validators)
    for key in ("block_height", "validators"):
        assert key in validators


@pytest.mark.asyncio
async def test_get_peers(client):
    peers = await client.get_peers()
    pprint(peers)
    assert isinstance(peers, list)
    assert len(peers)
    for peer in peers:
        for key in ("version", "network", "moniker", "id", "capabilities"):
            assert key in peer


@pytest.mark.asyncio
async def test_token_list(client):
    tokens = await client.get_token_list()
    pprint(tokens)
    for token in tokens:
        for key in (
            "mintable",
            "name",
            "original_symbol",
            "owner",
            "symbol",
            "total_supply",
        ):
            assert key in token


@pytest.mark.asyncio
async def test_get_markets(client):
    markets = await client.get_markets()
    pprint(markets)
    for market in markets:
        for key in (
            "base_asset_symbol",
            "list_price",
            "lot_size",
            "quote_asset_symbol",
            "tick_size",
        ):
            assert key in market


@pytest.mark.asyncio
async def test_get_fees(client):
    fees = await client.get_fees()
    assert fees
    assert "fee" in fees[0]


@pytest.mark.asyncio
async def test_get_transaction(client):
    h = "F9016F01A1098BF8024C28C8400AE010FC32DC8A393ADB26E56F37BC8B0C5D66"
    tx = await client.get_transaction(h)
    for key in ("ok", "log", "hash", "data"):
        assert key in tx
    assert tx["ok"]


@pytest.mark.asyncio
async def test_get_account(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    account = await client.get_account(address)
    assert "address" in account


@pytest.mark.asyncio
async def test_get_account_sequence(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    account = await client.get_account_sequence(address)
    assert "sequence" in account


@pytest.mark.asyncio
async def test_get_closed_orders(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    orders = await client.get_closed_orders(address, limit=1)
    assert "order" in orders

    start = (datetime.utcnow() - timedelta(days=1)).timestamp()
    end = datetime.utcnow().timestamp()
    markets = await client.get_markets()
    market = markets[0]
    symbol = "{base_asset_symbol}_{quote_asset_symbol}".format_map(market)
    orders = await client.get_closed_orders(
        address,
        start=start,
        end=end,
        offset=1,
        side=1,
        status="Ack",
        symbol=symbol,
        total=1,
    )
    assert "order" in orders


@pytest.mark.asyncio
async def test_get_open_orders(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    markets = await client.get_markets()
    market = markets[0]
    symbol = "{base_asset_symbol}_{quote_asset_symbol}".format_map(market)
    orders = await client.get_open_orders(
        address, limit=1, offset=0, total=1, symbol=symbol
    )
    pprint(orders)
    assert "order" in orders


@pytest.mark.asyncio
async def test_get_transactions(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    transactions = await client.get_transactions(address=address)
    assert "tx" in transactions

    transactions = await client.get_transactions(
        address=address,
        height=999,
        start=11,
        end=999999,
        limit=1,
        offset=0,
        side="RECEIVE",
        tx_asset="test",
        tx_type="NEW_ORDER",
    )
    assert "tx" in transactions


@pytest.mark.asyncio
async def test_get_order(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    orders = await client.get_closed_orders(address)
    closed_order = orders["order"][0]
    order_id = closed_order["orderId"]
    order = await client.get_order(order_id)
    assert order == closed_order


"""
@pytest.mark.asyncio
async def test_broadcast(client):
    body = ...
    tx = await client.broadcast(body)
    pprint(tx)

"""


@pytest.mark.asyncio
async def test_get_klines(client):
    markets = await client.get_markets()
    market = markets[0]
    symbol = "{base_asset_symbol}_{quote_asset_symbol}".format_map(market)
    interval = "1m"
    pprint(market)
    start = (datetime.utcnow() - timedelta(days=7)).timestamp()
    end = datetime.utcnow().timestamp()
    klines = await client.get_klines(
        symbol=symbol, interval=interval, start=start, end=end
    )
    pprint(klines)
    for kline in klines:
        assert len(kline) == 9


@pytest.mark.asyncio
async def test_get_ticker(client):
    tickers = await client.get_ticker()
    pprint(tickers)
    assert tickers
    for ticker in tickers:
        for key in ("askPrice", "askQuantity", "bidPrice", "bidQuantity"):
            assert key in ticker

    markets = await client.get_markets()
    market = markets[0]
    symbol = f"{market['base_asset_symbol']}_{market['quote_asset_symbol']}"
    tickers = await client.get_ticker(symbol=symbol)
    assert tickers
    for ticker in tickers:
        for key in ("askPrice", "askQuantity", "bidPrice", "bidQuantity"):
            assert key in ticker


@pytest.mark.asyncio
async def test_trades(client):
    trades = await client.get_trades()
    pprint(trades)
    assert "total" in trades
    assert "trade" in trades
    for trade in trades["trade"]:
        for key in (
            "baseAsset",
            "blockHeight",
            "buyFee",
            "buyerId",
            "buyerOrderId",
            "price",
            "quantity",
            "quoteAsset",
            "sellFee",
            "sellerId",
            "sellerOrderId",
            "symbol",
            "time",
            "tradeId",
        ):
            assert key in trade


@pytest.mark.asyncio
async def test_get_trades_for_address(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    trades = await client.get_trades(address=address)
    pprint(trades)
    assert "total" in trades
    assert "trade" in trades
    for trade in trades["trade"]:
        for key in (
            "baseAsset",
            "blockHeight",
            "buyFee",
            "buyerId",
            "buyerOrderId",
            "price",
            "quantity",
            "quoteAsset",
            "sellFee",
            "sellerId",
            "sellerOrderId",
            "symbol",
            "time",
            "tradeId",
        ):
            assert key in trade


@pytest.mark.asyncio
async def test_get_trades_other_params(client):
    trades = await client.get_trades(
        buyerOrderId="abc",
        height=999999,
        limit=1,
        offset=0,
        sellerOrderId="zyx",
        side=1,
        start=0,
        end=99999999,
        total=1,
        symbol="xyz",
    )
    assert trades


@pytest.mark.asyncio
async def test_get_block_exchange_fee(client):
    address = "tbnb18d6rnpmkxzqv3767xaaev5zzn06p42nya8zu79"
    fees = await client.get_block_exchange_fee(address)
    assert "blockExchangeFee" in fees
    fees = await client.get_block_exchange_fee(
        address, start=111111, end=99999999, offset=0, limit=1, total=1
    )
    assert "blockExchangeFee" in fees


@pytest.mark.asyncio
async def test_get_depth(client):
    markets = await client.get_markets()
    market = markets[0]
    symbol = f"{market['base_asset_symbol']}_{market['quote_asset_symbol']}"
    depth = await client.get_depth(symbol)
    assert "asks" in depth and "bids" in depth


@pytest.mark.asyncio
async def test_invalid_request(client):
    client._server = "https://binance.com/"
    try:
        resp = await client.post_request("INVALID")
        assert False, resp
    except BinanceChainException as e:
        assert e.response.status == 404
        assert isinstance(e.__cause__, aiohttp.ContentTypeError)


@pytest.mark.asyncio
async def test_del_without_close_warning():
    client = HTTPClient(testnet=True)
    await client.get_time()
    with pytest.warns(UserWarning):
        del (client)



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance Chain Node RPC HTTP API

https://docs.binance.org/api-reference/node-rpc.html#node-rpc

Endpoints that require arguments:

/subscribe?query=_
/unsubscribe?query=_
/unsubscribe_all?
"""
import sys
import itertools
import warnings
from typing import Any, Optional, Callable

import asyncio
import aiohttp

from .exceptions import BinanceChainException

MAINNET_URL = "https://seed1.longevito.io:443/"
TESTNET_URL = "https://seed-pre-s3.binance.org/"


class NodeRPC:
    """ Binance Chain Node RPC HTTP API Client """

    def __init__(self, url: str = None, testnet: bool = True):
        """
        :param: url: binance chain node URL
        :param testnet: A boolean to enable testnet
        :param session: An optional HTTP session to use
        """
        if not url:
            self.url = TESTNET_URL if testnet else MAINNET_URL
        self._id = itertools.count()
        self._session: Optional[aiohttp.ClientSession] = None
        self._testnet = testnet
        self._keepalive_task: Optional[asyncio.Future] = None

    def __del__(self):
        if self._session and not self._session.closed:
            warnings.warn(f"{repr(self)}.close() was never awaited")

    async def _request(self, method: str, path: str, **kwargs):
        """
        :method: `get` or `post`
        :path: the remote endpoint to call
        :kwargs: Extra arguments to pass to the request, like `params` or `data`.
        """
        if not self._session:
            self._session = aiohttp.ClientSession()
        try:
            resp = None
            async with getattr(self._session, method)(
                self.url + path, **kwargs
            ) as resp:
                return await resp.json()
        except Exception as e:
            raise BinanceChainException(resp) from e

    async def get_request(self, path: str, params: dict = None) -> Any:
        return await self._request("get", path, params=params)

    async def post_request(self, method: str, *params) -> dict:
        payload = {
            "method": method,
            "params": params,
            "jsonrpc": "2.0",
            "id": str(next(self._id)),
        }
        if not self._session:
            self._session = aiohttp.ClientSession()
        try:
            async with self._session.post(self.url, json=payload) as resp:
                return await resp.json()
        except Exception as e:
            raise BinanceChainException(resp) from e

    async def get_abci_info(self) -> dict:
        """Get some info about the application."""
        return await self.get_request("abci_info")

    async def get_consensus_state(self) -> dict:
        """ConsensusState returns a concise summary of the consensus state.
        This is just a snapshot of consensus state, and it will not persist.
        """
        return await self.get_request("consensus_state")

    async def get_dump_consensus_state(self) -> dict:
        """ConsensusState returns a concise summary of the consensus state.
        This is just a snapshot of consensus state, and it will not persist.
        """
        return await self.get_request("dump_consensus_state")

    async def get_genesis(self) -> dict:
        """Get genesis file."""
        return await self.get_request("genesis")

    async def get_health(self) -> dict:
        """Get node health.
        Returns empty result on success, no response - in case of an error.
        """
        return await self.get_request("health")

    async def get_net_info(self) -> dict:
        """Get network info."""
        return await self.get_request("net_info")

    async def get_num_unconfirmed_txs(self) -> dict:
        """Get number of unconfirmed transactions."""
        return await self.get_request("num_unconfirmed_txs")

    async def get_status(self) -> dict:
        """Get Tendermint status including node info, pubkey, latest block
        hash, app hash, block height and time.
        """
        return await self.get_request("status")

    async def abci_query(
        self, path: str, data: str = None, height: str = "0", prove: bool = False
    ) -> dict:
        """Query the application for some information.

        Available Query Paths:

            /store/acc/key
            /tokens/info
            /tokens/list
            /dex/pairs
            /dex/orderbook
            /param/fees
        """
        return await self.post_request("abci_query", path, data, height, prove)

    async def block(self, height: Optional[int] = None) -> dict:
        """Query the block at a given height.

        :param height: height of blockchain
        """
        return await self.post_request("block", height)

    async def block_by_hash(self, hash: str) -> dict:
        """Query a block by it's hash.
        :param hash: the block hash
        """
        return await self.post_request("block_by_hash", hash)

    async def block_results(self, height: Optional[str] = None) -> dict:
        """Gets ABCIResults at a given height."""
        return await self.post_request("block_results", height)

    async def blockchain(self, min_height: str, max_height: str) -> dict:
        """Get block headers for minHeight <= height <= maxHeight.
        Block headers are returned in descending order (highest first).
        """
        return await self.post_request("blockchain", min_height, max_height)

    async def broadcast_tx_async(self, tx: str) -> dict:
        """
        This method just returns the transaction hash right away and there is no
        return from CheckTx or DeliverTx.
        """
        return await self.post_request("broadcast_tx_async", tx)

    async def broadcast_tx_sync(self, tx: str) -> dict:
        """
        The transaction will be broadcasted and returns with the response from CheckTx.
        """
        return await self.post_request("broadcast_tx_sync", tx)

    async def broadcast_tx_commit(self, tx: str) -> dict:
        """
        The transaction will be broadcasted and returns with the response from
        CheckTx and DeliverTx.

        If CheckTx or DeliverTx fail, no error will be returned, but the
        returned result will contain a non-OK ABCI code.
        """
        return await self.post_request("broadcast_tx_commit", tx)

    async def commit(self, height: Optional[str] = None) -> dict:
        """Get block commit at a given height.
        If no height is provided, it will fetch the commit for the latest block.
        """
        return await self.post_request("commit", height)

    async def consensus_params(self, height: Optional[str] = None) -> dict:
        """Get consensus params at a given height."""
        return await self.post_request("consensus_params", height)

    async def tx(self, hash: str, prove: bool = False) -> dict:
        """Allows you to query the transaction results."""
        return await self.post_request("tx", hash, prove)

    async def tx_search(
        self, query: str, prove: bool = False, page: int = 1, per_page: int = 30
    ) -> dict:
        """Allows you to query for multiple transactions results.
        You could search transaction by its index. It returns a list of
        transactions (maximum ?per_page entries) and the total count.
        """
        return await self.post_request(
            "tx_search", query, prove, str(page), str(per_page)
        )

    async def unconfirmed_txs(self, limit: int = None) -> dict:
        """Get unconfirmed transactions."""
        return await self.post_request("unconfirmed_txs", limit)

    async def validators(self, height: str = None) -> dict:
        """Get information on the validators"""
        return await self.post_request("validators", height)

    def subscribe(
        self, query: str, callback: Optional[Callable[[dict], None]] = None
    ) -> None:
        """Subscribe to events via WebSocket.

        See list of all possible events here:
        https://godoc.org/github.com/tendermint/tendermint/types#pkg-constants
        For complete query syntax, check out:
        https://godoc.org/github.com/tendermint/tendermint/libs/pubsub/query.
        """
        payload = {
            "method": "subscribe",
            "params": [query],
            "jsonrpc": "2.0",
            "id": str(next(self._id)),
        }
        asyncio.ensure_future(self.send(payload))

    def unsubscribe(self, query: str) -> None:
        """Unubscribe from events via WebSocket."""
        payload = {
            "method": "unsubscribe",
            "params": [query],
            "jsonrpc": "2.0",
            "id": str(next(self._id)),
        }
        asyncio.ensure_future(self.send(payload))

    def unsubscribe_all(self) -> None:
        """Unubscribe from all events via WebSocket."""
        payload = {
            "method": "unsubscribe_all",
            "params": [],
            "jsonrpc": "2.0",
            "id": str(next(self._id)),
        }
        asyncio.ensure_future(self.send(payload))

    def start(
        self,
        on_open: Optional[Callable[[], None]] = None,
        on_msg: Callable[[], None] = None,
        on_error: Optional[Callable[[dict], None]] = None,
        loop: asyncio.AbstractEventLoop = None,
        keepalive: bool = True,
    ) -> None:
        """The main blocking call to start the WebSocket connection."""
        loop = loop or asyncio.get_event_loop()
        return loop.run_until_complete(
            self.start_async(on_open, on_msg, on_error, keepalive)
        )

    async def start_async(
        self,
        on_open: Optional[Callable[[], None]] = None,
        on_msg: Callable[[dict], None] = None,
        on_error: Optional[Callable[[dict], None]] = None,
        keepalive: bool = True,
    ) -> None:
        """Processes all websocket messages.

        :param callback: The single callback to use for all websocket messages
        :param keepalive: Run a background keepAlive coroutine
        """
        if not self._session:
            self._session = aiohttp.ClientSession()
        ws_url = f"{self.url}/websocket"
        async with self._session.ws_connect(ws_url) as ws:
            self._ws = ws
            if on_open:
                on_open()

            # Schedule keepalive calls every 30 minutes
            if keepalive:
                self._keepalive_task = asyncio.ensure_future(self._auto_keepalive())

            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    try:
                        data = msg.json()
                    except Exception as e:
                        print(f"Unable to decode msg: {msg}", file=sys.stderr)
                        continue
                    if data:
                        if on_msg:
                            on_msg(data)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    print(msg, file=sys.stderr)
                    if on_error:
                        on_error(msg)
                    break

    async def send(self, data: dict) -> None:
        """Send data to the WebSocket"""
        if not self._ws:
            print("Error: Cannot send to uninitialized websocket", file=sys.stderr)
            return
        await self._ws.send_json(data)

    async def _auto_keepalive(self):
        while True:
            await asyncio.sleep(30 * 60)
            self.keepalive()

    def close(self) -> None:
        """Close the websocket session"""
        if self._session and not self._session.closed:
            asyncio.ensure_future(self._session.close())
        if self._keepalive_task:
            self._keepalive_task.cancel()



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
    Intergrate with different methods of signing
"""

from eth_keyfile import decode_keyfile_json, create_keyfile_json
from bitcoinlib import keys, mnemonic, encoding
from .crypto import from_path, get_address
from secp256k1 import PrivateKey, PublicKey


HDPATH = "44'/714'/0'/0/0"
TESTNET_PREFIX = "tbnb"
MAINET_PREFIX = "bnb"


class Wallet:
    @staticmethod
    def create_wallet(password: str = "", testnet: bool = False):
        """
        Create brand new wallet
        """
        root_key = keys.HDKey(passphrase=password)
        key = from_path(root_key=root_key, path=HDPATH)
        return Wallet(key=key, testnet=testnet)

    @staticmethod
    def create_keystore(password: str = "") -> dict:
        """
            Create Keystore object
        """
        m = mnemonic.Mnemonic()
        mnem = m.generate(256)
        root_key = keys.HDKey.from_seed(m.to_seed(mnem, password=password))
        key = from_path(root_key=root_key, path=HDPATH)
        return create_keyfile_json(
            private_key=key.private_byte, password=bytes(password, "utf-8")
        )

    @staticmethod
    def create_wallet_mnemonic(
        language: str = "english", password: str = "", testnet: bool = False
    ):
        """
        Create wallet with mnemonic in language
        """
        m = mnemonic.Mnemonic(language)
        mnem = m.generate(256)
        root_key = keys.HDKey.from_seed(m.to_seed(mnem, password=password))
        key = from_path(root_key=root_key, path=HDPATH)
        return Wallet(key=key, testnet=testnet)

    @staticmethod
    def wallet_from_keystore(keystore: dict, password: str = "", testnet: bool = False):
        "Recover Binance wallet from keystore"
        private_key = decode_keyfile_json(
            keystore, password=encoding.to_bytes(password)
        )
        key = keys.HDKey(private_key)
        return Wallet(key=key, testnet=testnet)

    @staticmethod
    def wallet_from_privatekey(
        privatekey: str, password: str = "", testnet: bool = False
    ):
        """Recover Binance Wallet from privatekey"""
        key = keys.HDKey(import_key=privatekey, passphrase=password)
        return Wallet(key=key, testnet=testnet)

    @staticmethod
    def wallet_from_mnemonic(words: str, password: str = "", testnet: bool = False):
        "Recover wallet from mnemonic"
        m = mnemonic.Mnemonic(language="english")
        root_key = keys.HDKey.from_seed(m.to_seed(words=words, password=password))
        key = from_path(root_key=root_key, path=HDPATH)
        return Wallet(key=key, testnet=testnet)

    # @staticmethod
    # def wallet_from_seed(seed, testnet=False):
    #     root_key = keys.HDKey.from_seed(seed)
    #     key = from_path(root_key=root_key, path=HDPATH)
    #     return Wallet(key=key, testnet=testnet, mnemonic=words)

    def __init__(self, key: str, testnet: bool = False):
        self.testnet = testnet
        self.prefix = TESTNET_PREFIX if testnet else MAINET_PREFIX
        self.key = key
        self.address = get_address(prefix=self.prefix, key=key)

    def get_address(self) -> str:
        """Return wallet's address"""
        return self.address

    def get_privatekey(self):
        """Return wallet's private key"""
        return self.key.private_hex

    def get_publickey(self):
        """Return wallet's public key"""
        return self.key.public_hex

    def sign(self, msg):
        """
            Sign a message with private key, Return signature
        """
        priv = PrivateKey(self.key.private_byte, raw=True)
        sig = priv.ecdsa_sign(msg)
        h = priv.ecdsa_serialize_compact(sig)
        return self.key.public_hex, encoding.to_hexstring(h)

    def verify_signature(self, msg, signature):
        """Verify message and signature if its from this wallet"""
        pub = PublicKey(self.key.public_byte, raw=True)
        sig = pub.ecdsa_deserialize_compact(encoding.to_bytes(signature))
        valid = pub.ecdsa_verify(msg, sig)
        return valid



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance DEX WebSocket Test Suite
"""
import asyncio

import pytest

from binancechain import HTTPClient, WebSocket


def on_error(msg):
    print(f'Error: {msg}')


@pytest.fixture
async def client():
    # If we create fresh websockets too fast it may error?
    await asyncio.sleep(1)
    client = WebSocket(testnet=True)
    yield client
    client.close()


@pytest.fixture
async def symbols():
    symbols = []
    rest = HTTPClient(testnet=True)
    markets = await rest.get_markets()
    for market in markets:
        symbol = f"{market['base_asset_symbol']}_{market['quote_asset_symbol']}"
        symbols.append(symbol)
    yield symbols
    await rest.close()


@pytest.mark.asyncio
async def test_open_close(client):
    """"Open then immediately close"""
    def on_open():
        print('opened')
        client.close()

    await client.start_async(on_open=on_open, on_error=on_error)
    print('closed')


@pytest.mark.asyncio
async def test_trades(client, symbols):
    print(symbols)
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_trades(symbols=symbols, callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'trades'


@pytest.mark.asyncio
async def test_market_diff(client, symbols):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_market_diff(symbols=symbols, callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'marketDiff'


@pytest.mark.asyncio
async def test_market_depth(client, symbols):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_market_depth(symbols=symbols, callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'marketDepth'


@pytest.mark.asyncio
async def test_kline(client, symbols):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_kline(interval='1m', symbols=symbols, callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'kline_1m'


@pytest.mark.asyncio
async def test_tickers(client, symbols):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_ticker(symbols=symbols, callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'ticker'


@pytest.mark.asyncio
async def test_all_tickers(client):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_all_tickers(callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'allTickers'


@pytest.mark.asyncio
async def test_mini_ticker(client, symbols):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_mini_ticker(symbols=symbols, callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'miniTicker'


@pytest.mark.asyncio
async def test_all_mini_ticker(client, symbols):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_all_mini_tickers(callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert result['stream'] == 'allMiniTickers'


@pytest.mark.asyncio
async def test_blockheight(client):
    results = []

    def callback(msg):
        results.append(msg)
        client.close()

    def on_open():
        client.subscribe_blockheight(callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)

    result = results[0]
    assert 'stream' in result


@pytest.mark.asyncio
async def test_keepalive(client):
    def on_open():
        client.keepalive()
        client.close()
    await client.start_async(on_open=on_open, on_error=on_error)


@pytest.mark.asyncio
async def test_unsubscribe(client):
    results = []

    def callback(msg):
        results.append(msg)
        client.unsubscribe("blockheight")
        client.close()

    def on_open():
        client.subscribe_blockheight(callback=callback)

    await client.start_async(on_open=on_open, on_error=on_error)
    assert results


@pytest.mark.asyncio
async def test_decorator(client):
    @client.on('open')
    def callback():
        client.close()
    await client.start_async()


@pytest.mark.asyncio
async def test_decorator_async(client):
    @client.on('open')
    async def callback():
        client.close()
    await client.start_async()


@pytest.mark.asyncio
async def test_decorator_sub_queue(client):
    results = []

    @client.on("allTickers", symbols=["$all"])
    async def callback(msg):
        results.append(msg)
        client.close()

    await client.start_async()
    assert results



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT

import aiohttp


class BinanceChainException(Exception):

    def __init__(self, response: aiohttp.ClientResponse = None):
        self.response = response

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.response}>"



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
    Create and manage wallets
"""
import binascii
from typing import Union, Any, Tuple, Optional
import json
from bitcoinlib import encoding
from varint import encode
from decimal import Decimal
from .crypto import generate_id, address_decode
from .httpclient import HTTPClient
from .enums import Ordertype, Side, Timeinforce, Votes
from .transaction_pb2 import (
    CancelOrder,
    Freeze,
    NewOrder,
    Send,
    StdSignature,
    Unfreeze,
    Vote,
    StdTx,
    Input,
    Output,
    Token,
)

SOURCE = "1"
TESTNET_CHAIN_ID = "Binance-Chain-Nile"
MAINNET_CHAIN_ID = "Binance-Chain-Tigris"
TYPE_PREFIX = {
    "CancelOrder": b"166E681B",
    "TokenFreeze": b"E774B32D",
    "TokenUnfreeze": b"6515FF0D",
    "NewOrder": b"CE6DC043",
    "Send": b"2A2C87FA",
    "PubKey": b"EB5AE987",
    "StdTx": b"F0625DEE",
    "Vote": b"A1CADD36",
}
MSG_TYPES = {
    "CancelOrder": CancelOrder,
    "TokenFreeze": Freeze,
    "TokenUnfreeze": Unfreeze,
    "NewOrder": NewOrder,
    "Send": Send,
    "Vote": Vote,
}

BASE = 100000000
number_type = Union[str, float, int, Decimal]


class TransactionBase:
    def __init__(
        self,
        address: str,
        account_number,
        sequence,
        chainid: str,
        memo: str = "",
        data: str = "",
    ):
        self.account_number = account_number
        self.data = data.encode()
        self.memo = memo
        self.address = address
        self.sequence = sequence
        self.StdSignMsg = {
            "memo": self.memo,
            "msgs": [dict],
            "account_number": str(self.account_number),
            "sequence": str(self.sequence),
            "chain_id": chainid,
            "source": str(SOURCE),
            "data": None,
        }
        print("CHAIN ID ", chainid)

    def get_new_order_msg(
        self,
        symbol: str,
        side: Side,
        price: number_type,
        quantity: number_type,
        sequence=None,
        ordertype: Ordertype = Ordertype.LIMIT,
        timeInForce: Timeinforce = Timeinforce.GTE,
    ):
        """Create new_order protobuf attributes, SignMessage of the transaction"""
        id = generate_id(self.address, self.sequence)
        price = int(Decimal(price) * BASE)
        quantity = int(Decimal(quantity) * Decimal(100000000))
        self.msg = {
            "symbol": symbol,
            "sender": self.address,
            "id": id,
            "side": side.value,
            "ordertype": ordertype.value,
            "timeinforce": timeInForce.value,
            "price": price,
            "quantity": quantity,
        }
        self.StdSignMsg["msgs"] = [self.msg]
        self.SignMessage = json.dumps(
            self.StdSignMsg, sort_keys=True, separators=(",", ":")
        ).encode()
        self.stdMsg = self.generate_stdNewOrderMsg(self.msg)
        return self.SignMessage

    def generate_stdNewOrderMsg(self, msg: dict) -> bytes:
        """Generate StdMsg part of StdTx"""
        std = NewOrder()
        std.sender = address_decode(self.address)
        std.id = generate_id(self.address, self.sequence)
        std.ordertype = msg[
            "ordertype"
        ]  # currently only 1 type : limit =2, will change in the future
        std.symbol = msg["symbol"].encode()
        std.side = msg["side"]
        std.price = msg["price"]
        std.quantity = msg["quantity"]
        std.timeinforce = msg["timeinforce"]
        proto_bytes = std.SerializeToString()
        type_bytes = encoding.to_bytes(TYPE_PREFIX["NewOrder"])
        return type_bytes + proto_bytes

    def get_cancel_order_msg(self, symbol: str, refid: str):
        """Generate cancel_order StdMsg for StdTx and SignMessage for current transaction"""
        self.msg = {"sender": self.address, "symbol": symbol, "refid": refid}
        self.StdSignMsg["msgs"] = [self.msg]
        self.SignMessage = json.dumps(
            self.StdSignMsg, sort_keys=True, separators=(",", ":")
        ).encode()
        std = CancelOrder()
        std.symbol = symbol
        std.sender = address_decode(self.address)
        std.refid = refid
        self.stdMsg = (
            encoding.to_bytes(TYPE_PREFIX["CancelOrder"]) + std.SerializeToString()
        )
        return self.SignMessage

    def get_transfer_msg(self, to_address: str, symbol: str, amount: number_type):
        """Generate transfer StdMsg for StdTx and SignMessage for current transaction"""
        amount = int(Decimal(amount) * BASE)
        self.msg = {
            "inputs": [
                {
                    "address": self.address,
                    "coins": [{"denom": symbol, "amount": amount}],
                }
            ],
            "outputs": [
                {"address": to_address, "coins": [{"denom": symbol, "amount": amount}]}
            ],
        }
        self.StdSignMsg["msgs"] = [self.msg]

        self.SignMessage = json.dumps(
            self.StdSignMsg, sort_keys=True, separators=(",", ":")
        ).encode()
        std = Send()
        input = Input()
        output = Output()
        token_proto = Token()
        token_proto.amount = amount
        token_proto.denom = symbol.encode()
        input.address = address_decode(self.address)
        input.coins.extend([token_proto])
        output.address = address_decode(to_address)
        output.coins.extend([token_proto])
        std.inputs.extend([input])
        std.outputs.extend([output])
        self.stdMsg = encoding.to_bytes(TYPE_PREFIX["Send"]) + std.SerializeToString()
        return self.SignMessage

    def get_freeze_token_msg(self, symbol: str, amount: number_type):
        """Generate freeze_token StdMsg for StdTx and SignMessage for current transaction"""
        amount = int(Decimal(amount) * BASE)
        self.msg = {"from": self.address, "symbol": symbol, "amount": amount}
        self.StdSignMsg["msgs"] = [self.msg]
        self.SignMessage = json.dumps(
            self.StdSignMsg, sort_keys=True, separators=(",", ":")
        ).encode()
        std = Freeze()
        setattr(std, "from", address_decode(self.address))
        std.symbol = symbol
        std.amount = amount
        self.stdMsg = (
            encoding.to_bytes(TYPE_PREFIX["TokenFreeze"]) + std.SerializeToString()
        )
        return self.SignMessage

    def get_unfreeze_token_msg(self, symbol: str, amount: number_type):
        """Generate unfreeze_token StdMsg for StdTx and SignMessage for current transaction"""
        amount = int(Decimal(amount) * BASE)
        self.msg = {"from": self.address, "symbol": symbol, "amount": amount}
        self.StdSignMsg["msgs"] = [self.msg]
        self.SignMessage = json.dumps(
            self.StdSignMsg, sort_keys=True, separators=(",", ":")
        ).encode()
        std = Freeze()
        setattr(std, "from", address_decode(self.address))
        std.symbol = symbol
        std.amount = amount
        self.stdMsg = (
            encoding.to_bytes(TYPE_PREFIX["TokenUnfreeze"]) + std.SerializeToString()
        )
        return self.SignMessage

    def get_vote_msg(self, proposal_id: Union[str, int], option: Votes):
        """Generate cancel_order StdMsg for StdTx and SignMessage for current transaction"""
        self.msg = {
            "proposal_id": proposal_id,
            "voter": self.address,
            "option": option.value,
        }
        self.StdSignMsg["msgs"] = [self.msg]
        self.SignMessage = json.dumps(
            self.StdSignMsg, sort_keys=True, separators=(",", ":")
        ).encode()
        std = Vote()
        std.voter = address_decode(self.address)
        std.proposal_id = proposal_id
        std.option = option.value
        self.stdMsg = encoding.to_bytes(TYPE_PREFIX["Vote"]) + std.SerializeToString()
        return self.SignMessage

    def get_sign_message(self):
        return self.SignMessage

    def update_signature(self, pubkey: str, signature: str):
        """
        :Update current transaction with pubkey and signature from self.address
        :Create StdTx proto
        :Return data hex ready to be broadcast
        """
        pubkey_bytes = self.pubkey_to_msg(pubkey)
        self.stdSignature = self.generate_stdSignatureMsg(pubkey_bytes, signature)
        self.stdTx = self.generate_StdTxMsg()
        return binascii.hexlify(self.stdTx)

    def pubkey_to_msg(self, pubkey: str):
        key_bytes = encoding.to_bytes(pubkey)
        return (
            encoding.to_bytes(TYPE_PREFIX["PubKey"])
            + encode(len(key_bytes))
            + key_bytes
        )

    def generate_stdSignatureMsg(self, pubkey_bytes: bytes, signature: str):
        """Generate StdSignature for StdTx"""
        std = StdSignature()
        std.pub_key = pubkey_bytes
        std.signature = encoding.to_bytes(signature)
        std.account_number = self.account_number
        std.sequence = self.sequence
        proto_bytes = std.SerializeToString()
        return proto_bytes

    def generate_StdTxMsg(self):
        """Geneate StdTx"""
        std = StdTx()
        std.msgs.extend([self.stdMsg])
        std.signatures.extend([self.stdSignature])
        std.memo = self.memo
        std.source = 1
        std.data = self.data
        proto_bytes = std.SerializeToString()
        type_bytes = encoding.to_bytes(TYPE_PREFIX["StdTx"])
        return encode(len(proto_bytes) + len(type_bytes)) + type_bytes + proto_bytes

    def ___repr__(self):
        return "test string transaction"


class Transaction:
    @staticmethod
    async def new_order_transaction(
        address: str,
        symbol: str,
        side: Side,
        price: number_type,
        quantity: number_type,
        ordertype: Ordertype = Ordertype.LIMIT,
        timeInForce: Timeinforce = Timeinforce.GTE,
        testnet: bool = False,
        account_number: int = None,
        sequence: int = None,
        client: Any = None,
    ) -> TransactionBase:
        """ Create New Order TransactionBase object"""
        if not client:
            client = HTTPClient(testnet=testnet)
            chain_id = TESTNET_CHAIN_ID if testnet else MAINNET_CHAIN_ID
        else:
            chain_id = TESTNET_CHAIN_ID if client._testnet else MAINNET_CHAIN_ID
        if not account_number or not sequence:
            account_info = await client.get_account(address)
            account_number = account_info["account_number"]
            sequence = account_info["sequence"]
        transaction = TransactionBase(
            address, account_number=account_number, sequence=sequence, chainid=chain_id
        )
        transaction.get_new_order_msg(
            symbol=symbol,
            side=side,
            ordertype=ordertype,
            price=price,
            quantity=quantity,
            timeInForce=timeInForce,
        )
        return transaction

    @staticmethod
    async def cancel_order_transaction(
        address: str,
        symbol: str,
        refid: str,
        testnet: bool = False,
        account_number: int = None,
        sequence: int = None,
        client: Any = None,
    ) -> TransactionBase:
        """
            Create Cancel order TransactionBase object
        """
        if not client:
            client = HTTPClient(testnet=testnet)
            chain_id = TESTNET_CHAIN_ID if testnet else MAINNET_CHAIN_ID
        else:
            chain_id = TESTNET_CHAIN_ID if client._testnet else MAINNET_CHAIN_ID
        if not account_number or not sequence:
            account_info = await client.get_account(address)
            account_number = account_info["account_number"]
            sequence = account_info["sequence"]
        transaction = TransactionBase(
            address, account_number=account_number, sequence=sequence, chainid=chain_id
        )
        transaction.get_cancel_order_msg(symbol=symbol, refid=refid)
        return transaction

    @staticmethod
    async def transfer_transaction(
        from_address: str,
        to_address: str,
        symbol: str,
        amount: number_type,
        testnet: bool = False,
        account_number: int = None,
        sequence: int = None,
        client: Any = None,
    ) -> TransactionBase:
        """
            Create transfer Transaction Base object
        """
        if not client:
            client = HTTPClient(testnet=testnet)
            chain_id = TESTNET_CHAIN_ID if testnet else MAINNET_CHAIN_ID
        else:
            chain_id = TESTNET_CHAIN_ID if client._testnet else MAINNET_CHAIN_ID
        if not account_number or not sequence:
            account_info = await client.get_account(from_address)
            account_number = account_info["account_number"]
            sequence = account_info["sequence"]
        transaction = TransactionBase(
            address=from_address,
            account_number=account_number,
            sequence=sequence,
            chainid=chain_id,
        )
        transaction.get_transfer_msg(
            to_address=to_address, symbol=symbol, amount=amount
        )
        return transaction

    @staticmethod
    async def freeze_token_transaction(
        address: str,
        symbol: str,
        amount: number_type,
        testnet: bool = False,
        account_number: int = None,
        sequence: int = None,
        client: Any = None,
    ) -> TransactionBase:
        """
        Create free_token TransactionBase object
        """
        if not client:
            client = HTTPClient(testnet=testnet)
            chain_id = TESTNET_CHAIN_ID if testnet else MAINNET_CHAIN_ID
        else:
            chain_id = TESTNET_CHAIN_ID if client._testnet else MAINNET_CHAIN_ID
        if not account_number or not sequence:
            account_info = await client.get_account(address)
            account_number = account_info["account_number"]
            sequence = account_info["sequence"]
        transaction = TransactionBase(
            address=address,
            account_number=account_number,
            sequence=sequence,
            chainid=chain_id,
        )
        transaction.get_freeze_token_msg(symbol=symbol, amount=amount)
        return transaction

    @staticmethod
    async def unfreeze_token_transaction(
        address: str,
        symbol: str,
        amount: number_type,
        testnet: bool = False,
        account_number: int = None,
        sequence: int = None,
        client: Any = None,
    ) -> TransactionBase:
        """
        Create unfreeze token TransactionBase object
        """
        if not client:
            client = HTTPClient(testnet=testnet)
            chain_id = TESTNET_CHAIN_ID if testnet else MAINNET_CHAIN_ID
        else:
            chain_id = TESTNET_CHAIN_ID if client._testnet else MAINNET_CHAIN_ID
        if not account_number or not sequence:
            account_info = await client.get_account(address)
            account_number = account_info["account_number"]
            sequence = account_info["sequence"]
        transaction = TransactionBase(
            address=address,
            account_number=account_number,
            sequence=sequence,
            chainid=chain_id,
        )
        transaction.get_unfreeze_token_msg(symbol=symbol, amount=amount)
        return transaction

    @staticmethod
    async def vote_transaction(
        voter: str,
        proposal_id: Union[int, str],
        option: Votes,
        testnet: bool = False,
        account_number: int = None,
        sequence: int = None,
        client: Any = None,
    ) -> TransactionBase:
        """
        Create vote TransactionBase object
        """
        if not client:
            client = HTTPClient(testnet=testnet)
            chain_id = TESTNET_CHAIN_ID if testnet else MAINNET_CHAIN_ID
        else:
            chain_id = TESTNET_CHAIN_ID if client._testnet else MAINNET_CHAIN_ID
        if not account_number or not sequence:
            account_info = await client.get_account(voter)
            account_number = account_info["account_number"]
            sequence = account_info["sequence"]
        transaction = TransactionBase(
            address=voter,
            account_number=account_number,
            sequence=sequence,
            chainid=chain_id,
        )
        transaction.get_vote_msg(proposal_id=proposal_id, option=option)
        return transaction

    def __init__(self, wallet: Any, client: Any = None, testnet: bool = False):
        self.wallet = wallet
        self.address = wallet.get_address()
        if not client:
            self.client = HTTPClient(testnet=testnet)
        else:
            self.client = client

    async def get_account_info(self) -> Tuple[int, int]:
        """Get account number and current valid sequence number"""
        account_info = await self.client.get_account(self.address)
        account_number = account_info["account_number"]
        sequence = account_info["sequence"]
        return account_number, sequence

    async def create_new_order(
        self,
        symbol: str,
        side: Side,
        ordertype: Ordertype,
        price: number_type,
        quantity: number_type,
        timeInForce: Timeinforce,
    ) -> Any:
        """
            Create,sign and broadcast new_order tranasction
        """
        account_number, sequence = await self.get_account_info()
        transaction = await Transaction.new_order_transaction(
            address=self.address,
            client=self.client,
            account_number=account_number,
            sequence=sequence,
            symbol=symbol,
            side=side,
            ordertype=ordertype,
            price=price,
            quantity=quantity,
            timeInForce=timeInForce,
        )
        return await self.sign_and_broadcast(transaction)

    async def cancel_order(self, symbol: str, refid: str) -> Any:
        """Create, sign and broadcast cancel_order transaction"""
        account_number, sequence = await self.get_account_info()
        transaction = await Transaction.cancel_order_transaction(
            address=self.address,
            client=self.client,
            account_number=account_number,
            sequence=sequence,
            symbol=symbol,
            refid=refid,
        )
        return await self.sign_and_broadcast(transaction)

    async def transfer(self, to_address: str, symbol: str, amount: number_type) -> Any:
        """Create, sign and broadcast transfer transaction"""
        account_number, sequence = await self.get_account_info()
        transaction = await Transaction.transfer_transaction(
            from_address=self.address,
            client=self.client,
            account_number=account_number,
            sequence=sequence,
            to_address=to_address,
            symbol=symbol,
            amount=amount,
        )
        return await self.sign_and_broadcast(transaction)

    async def freeze_token(self, symbol: str, amount: number_type) -> Any:
        """Create, sign and broadcast free_token transaction"""
        account_number, sequence = await self.get_account_info()
        transaction = await Transaction.freeze_token_transaction(
            address=self.address,
            client=self.client,
            account_number=account_number,
            sequence=sequence,
            symbol=symbol,
            amount=amount,
        )
        return await self.sign_and_broadcast(transaction)

    async def unfreeze_token(self, symbol: str, amount: number_type) -> Any:
        """Create, sign and broadcast unfreeze_token transaction"""
        account_number, sequence = await self.get_account_info()
        transaction = await Transaction.unfreeze_token_transaction(
            address=self.address,
            client=self.client,
            account_number=account_number,
            sequence=sequence,
            symbol=symbol,
            amount=amount,
        )
        return await self.sign_and_broadcast(transaction)

    async def vote(self, proposal_id: str, option: Votes):
        account_number, sequence = await self.get_account_info()
        transaction = await Transaction.vote_transaction(
            voter=self.address,
            client=self.client,
            account_number=account_number,
            sequence=sequence,
            proposal_id=proposal_id,
            option=option,
        )
        return await self.sign_and_broadcast(transaction)

    async def sign_and_broadcast(self, transaction: TransactionBase) -> Any:
        """Sign and broadcast an TransactionBase object"""
        pub, sig = self.wallet.sign(transaction.get_sign_message())
        hex_data = transaction.update_signature(pub, sig)
        broadcast_info = await self.client.broadcast(hex_data)
        return broadcast_info



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT

from bitcoinlib import encoding
import bech32


def from_path(root_key, path):
    if isinstance(path, str):
        p = path.rstrip("/").split("/")
    elif isinstance(path, bytes):
        p = path.decode("utf-8").rstrip("/").split("/")
    else:
        p = list(path)
    key = root_key
    for i in p:
        if isinstance(i, str):
            hardened = i[-1] == "'"
            index = int(i[:-1], 0) | 0x80000000 if hardened else int(i, 0)
        else:
            index = i
        key = key.child_private(index, hardened=hardened)
    return key


def get_address(prefix, key):
    convert = encoding.convertbits(encoding.hash160(key.public_compressed_byte), 8, 5)
    address = bech32.bech32_encode(prefix, convert)
    return address


def address_decode(address):
    prefix, words = bech32.bech32_decode(address)
    convert = encoding.convertbits(words, 5, 8)
    return bytes(convert)


def generate_signature(key, data):
    pass


def verify_signature(data, signed_data, pubkey):
    pass


def generate_id(address, sequence):
    prefix, words = bech32.bech32_decode(address)
    convert_w = encoding.convertbits(words, 5, 8)
    decodedAddress = encoding.to_bytearray(convert_w).hex()
    return f"{decodedAddress.upper()}-{sequence+1}"



from .enums import Ordertype, Side, Votes, Timeinforce
from .httpclient import HTTPClient
from .noderpc import NodeRPC
from .transaction import Transaction
from .wallet import Wallet
from .websocket import WebSocket
from .exceptions import BinanceChainException



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance DEX WebSockets

https://docs.binance.org/api-reference/dex-api/ws-streams.html#websocket-streams
"""
import asyncio
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import aiohttp
from pyee import AsyncIOEventEmitter

MAINNET_URL = "wss://dex.binance.org/api/ws"
TESTNET_URL = "wss://testnet-dex.binance.org/api/ws"


class WebSocket:
    """The Binance DEX WebSocket Manager."""

    def __init__(
        self,
        address: str = None,
        testnet: bool = False,
        keepalive: bool = True,
        loop: asyncio.AbstractEventLoop = None,
        url: str = None,
    ) -> None:
        if not url:
            self.url = TESTNET_URL if testnet else MAINNET_URL
        else:
            self.url = url
        self.address = address
        self._session = aiohttp.ClientSession()
        self._ws: Optional[aiohttp.ClientWebSocketResponse] = None
        self._loop = loop or asyncio.get_event_loop()
        self._events = AsyncIOEventEmitter(loop=self._loop)
        self._sub_queue: List[Tuple[str, dict]] = []
        self._keepalive = keepalive
        self._keepalive_task: Optional[asyncio.Future] = None
        self._open = False
        self._testnet = testnet

    def on(self, event: str, func: Optional[Callable] = None, **kwargs):
        """Register an event, and optional handler.

        This can be used as a decorator or as a normal method.
        See `examples/websockets_decorator.py` for usage.
        """
        # Queue up most events from startup-time decorators until after we are open
        if not self._open and event not in ("open", "error", "new_listener"):
            self._sub_queue.append((event, kwargs))
        if func:
            self._events.on(event, func)
            return None
        else:
            return self._events.on(event)

    def start(
        self,
        on_open: Optional[Callable[[], None]] = None,
        on_error: Optional[Callable[[dict], None]] = None,
        loop: asyncio.AbstractEventLoop = None,
    ) -> None:
        """The main blocking call to start the WebSocket connection."""
        loop = loop or asyncio.get_event_loop()
        return loop.run_until_complete(self.start_async(on_open, on_error))

    async def start_async(
        self,
        on_open: Optional[Callable[[], None]] = None,
        on_error: Optional[Callable[[dict], None]] = None,
    ) -> None:
        """Processes all websocket messages."""
        if self.address:  # address-specific socket
            url = f"{self.url}/{self.address}"
        else:
            url = self.url

        async with self._session.ws_connect(url) as ws:
            self._ws = ws
            self._events.emit("open")
            while self._sub_queue:
                event, kwargs = self._sub_queue.pop()
                self.subscribe(event, **kwargs)
            if on_open:
                on_open()

            # Schedule keepalive calls every 30 minutes
            if self._keepalive:
                self._keepalive_task = asyncio.ensure_future(self._auto_keepalive())

            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    try:
                        data = msg.json()
                    except Exception as e:
                        print(f"Unable to decode msg: {msg}", file=sys.stderr)
                        continue
                    if not data:
                        print(f"Got empty msg: {msg}", file=sys.stderr)
                        continue
                    if "error" in data:
                        self._events.emit("error", data)
                        if on_error:
                            on_error(data)
                        else:
                            print(f"Unhandled error msg: {data}", file=sys.stderr)
                        continue
                    if "stream" not in data:
                        print(f"Got msg without stream: {data}", file=sys.stderr)
                        continue
                    if "data" not in data:
                        print(f"Got msg without data: {data}", file=sys.stderr)
                        continue

                    self._events.emit(data["stream"], data)

                elif msg.type == aiohttp.WSMsgType.ERROR:
                    print(msg, file=sys.stderr)
                    self._events.emit("error", msg)
                    break

    async def send(self, data: dict) -> None:
        """Send data to the WebSocket"""
        if not self._ws:
            print("Error: Cannot send to uninitialized websocket", file=sys.stderr)
            return
        await self._ws.send_json(data)

    def subscribe(
        self,
        stream: str,
        symbols: Optional[List[str]] = None,
        address: Optional[str] = None,
        callback: Optional[Callable[[dict], None]] = None,
    ):
        """Subscribe to a WebSocket stream.

        See the documentation for more details on the available streams
        https://docs.binance.org/api-reference/dex-api/ws-streams.html
        """
        payload: Dict[Any, Any] = {"method": "subscribe", "topic": stream}
        if symbols:
            payload["symbols"] = symbols
        if address:
            payload["address"] = address
        elif self.address:
            payload["address"] = self.address
        self._events.on(stream, callback)
        asyncio.ensure_future(self.send(payload))

    def unsubscribe(self, stream, symbols=None) -> None:
        payload = {"method": "unsubscribe", "topic": stream}
        if symbols:
            payload["symbols"] = symbols
        asyncio.ensure_future(self.send(payload))

    def subscribe_user_orders(
        self, callback: Callable[[dict], None], address: Optional[str] = None
    ) -> None:
        """Subscribe to individual order updates."""
        self.subscribe("orders", address=address, callback=callback)

    def subscribe_user_accounts(
        self, callback: Callable[[dict], None], address: Optional[str] = None
    ) -> None:
        """Subscribe to account updates."""
        self.subscribe("accounts", address=address, callback=callback)

    def subscribe_user_transfers(
        self, callback: Callable[[dict], None], address: Optional[str] = None
    ) -> None:
        """
        Subscribe to transfer updates if `address` is involved (as sender or
        receiver) in a transfer. Multisend is also covered.
        """
        self.subscribe("transfers", address=address, callback=callback)

    def subscribe_trades(
        self, symbols: List[str], callback: Callable[[dict], None]
    ) -> None:
        """Subscribe to individual trade updates."""
        self.subscribe("trades", symbols=symbols, callback=callback)

    def subscribe_market_diff(
        self, symbols: List[str], callback: Callable[[dict], None]
    ) -> None:
        "Order book price and quantity depth updates used to locally keep an order book." ""
        self.subscribe("marketDiff", symbols=symbols, callback=callback)

    def subscribe_market_depth(
        self, symbols: List[str], callback: Callable[[dict], None]
    ) -> None:
        """Top 20 levels of bids and asks."""
        self.subscribe("marketDepth", symbols=symbols, callback=callback)

    def subscribe_kline(
        self, interval: str, symbols: List[str], callback: Callable[[dict], None]
    ) -> None:
        """
        The kline/candlestick stream pushes updates to the current
        klines/candlestick every second.

        Kline/Candlestick chart intervals:
            m -> minutes; h -> hours; d -> days; w -> weeks; M -> months
            1m 3m 5m 15m 30m 1h 2h 4h 6h 8h 12h 1d 3d 1w 1M
        """
        self.subscribe(f"kline_{interval}", symbols=symbols, callback=callback)

    def subscribe_ticker(
        self, symbols: List[str], callback: Callable[[dict], None]
    ) -> None:
        """24hr Ticker statistics for a single symbol are pushed every second."""
        self.subscribe("ticker", symbols=symbols, callback=callback)

    def subscribe_all_tickers(self, callback: Callable[[dict], None]) -> None:
        """24hr Ticker statistics for a all symbols are pushed every second."""
        self.subscribe("allTickers", symbols=["$all"], callback=callback)

    def subscribe_mini_ticker(
        self, symbols: List[str], callback: Callable[[dict], None]
    ) -> None:
        """A ticker for a single symbol is pushed every second."""
        self.subscribe("miniTicker", symbols=symbols, callback=callback)

    def subscribe_all_mini_tickers(self, callback: Callable[[dict], None]) -> None:
        """Array of 24hr Mini Ticker statistics for a all symbols pushed every second."""
        self.subscribe("allMiniTickers", symbols=["$all"], callback=callback)

    def subscribe_blockheight(self, callback: Callable[[dict], None]) -> None:
        """Streams the latest block height."""
        self.subscribe("blockheight", symbols=["$all"], callback=callback)

    def keepalive(self) -> None:
        """Extend the connection time by another 30 minutes"""
        asyncio.ensure_future(self.send({"method": "keepAlive"}))

    async def _auto_keepalive(self):
        while True:
            await asyncio.sleep(30 * 60)
            self.keepalive()

    def close(self) -> None:
        """Close the websocket session"""
        asyncio.ensure_future(self.send({"method": "close"}))
        if self._session:
            asyncio.ensure_future(self._session.close())
        if self._keepalive_task:
            self._keepalive_task.cancel()



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance Chain Node RPC Test Suite
"""
import pytest
import aiohttp

from binancechain import NodeRPC, BinanceChainException


@pytest.fixture
async def noderpc():
    noderpc = NodeRPC(testnet=True)
    yield noderpc
    noderpc.close()


def on_error(msg):
    print(f'Error: {msg}')


@pytest.mark.asyncio
async def test_abci_info(noderpc):
    resp = await noderpc.get_abci_info()
    assert "result" in resp


@pytest.mark.asyncio
async def test_consensus_state(noderpc):
    resp = await noderpc.get_consensus_state()
    assert "result" in resp


@pytest.mark.asyncio
async def test_get_dump_consensus_state(noderpc):
    resp = await noderpc.get_dump_consensus_state()
    assert "result" in resp


@pytest.mark.asyncio
async def test_get_genesis(noderpc):
    resp = await noderpc.get_genesis()
    assert "result" in resp


@pytest.mark.asyncio
async def test_get_health(noderpc):
    resp = await noderpc.get_health()
    assert "result" in resp


@pytest.mark.asyncio
async def test_net_info(noderpc):
    resp = await noderpc.get_net_info()
    assert "result" in resp


@pytest.mark.asyncio
async def test_get_num_unconfirmed_txs(noderpc):
    resp = await noderpc.get_num_unconfirmed_txs()
    assert "result" in resp


@pytest.mark.asyncio
async def test_get_status(noderpc):
    resp = await noderpc.get_status()
    assert "result" in resp


@pytest.mark.asyncio
async def test_abci_query(noderpc):
    resp = await noderpc.abci_query("/param/fees")
    assert "result" in resp


@pytest.mark.asyncio
async def test_block(noderpc):
    resp = await noderpc.block()
    assert "result" in resp


@pytest.mark.asyncio
async def test_block_by_hash(noderpc):
    resp = await noderpc.block_by_hash(
        "8B43B22699803E8F8E51E7D9294C14EDAD34A264ECFB7F776AF9422714B93C06"
    )
    assert "error" in resp


@pytest.mark.asyncio
async def test_block_results(noderpc):
    resp = await noderpc.block_results()
    assert "result" in resp


@pytest.mark.asyncio
async def test_blockchain(noderpc):
    resp = await noderpc.blockchain("0", "1")
    assert "result" in resp


@pytest.mark.asyncio
async def test_broadcast_tx_async(noderpc):
    resp = await noderpc.broadcast_tx_async(
        "0xdb01f0625dee0a63ce6dc0430a14813e4939f1567b219704ffc2ad4df58bde010879122b383133453439333946313536374232313937303446464332414434444635384244453031303837392d34341a0d5a454252412d3136445f424e422002280130c0843d38904e400112700a26eb5ae9872102139bdd95de72c22ac2a2b0f87853b1cca2e8adf9c58a4a689c75d3263013441a124015e99f7a686529c76ccc2d70b404af82ca88dfee27c363439b91ea0280571b2731c03b902193d6a5793baf64b54bcdf3f85e0d7cf657e1a1077f88143a5a65f518d2e518202b"
    )
    assert "result" in resp


@pytest.mark.asyncio
async def test_broadcast_tx_commit(noderpc):
    resp = await noderpc.broadcast_tx_commit(
        "0xdb01f0625dee0a63ce6dc0430a14813e4939f1567b219704ffc2ad4df58bde010879122b383133453439333946313536374232313937303446464332414434444635384244453031303837392d34341a0d5a454252412d3136445f424e422002280130c0843d38904e400112700a26eb5ae9872102139bdd95de72c22ac2a2b0f87853b1cca2e8adf9c58a4a689c75d3263013441a124015e99f7a686529c76ccc2d70b404af82ca88dfee27c363439b91ea0280571b2731c03b902193d6a5793baf64b54bcdf3f85e0d7cf657e1a1077f88143a5a65f518d2e518202b"
    )
    assert "result" in resp


@pytest.mark.asyncio
async def test_broadcast_tx_sync(noderpc):
    resp = await noderpc.broadcast_tx_sync(
        "0xdb01f0625dee0a63ce6dc0430a14813e4939f1567b219704ffc2ad4df58bde010879122b383133453439333946313536374232313937303446464332414434444635384244453031303837392d34341a0d5a454252412d3136445f424e422002280130c0843d38904e400112700a26eb5ae9872102139bdd95de72c22ac2a2b0f87853b1cca2e8adf9c58a4a689c75d3263013441a124015e99f7a686529c76ccc2d70b404af82ca88dfee27c363439b91ea0280571b2731c03b902193d6a5793baf64b54bcdf3f85e0d7cf657e1a1077f88143a5a65f518d2e518202b"
    )
    assert "result" in resp


@pytest.mark.asyncio
async def test_commit(noderpc):
    resp = await noderpc.commit("1")
    assert "result" in resp


@pytest.mark.asyncio
async def test_consensus_params(noderpc):
    resp = await noderpc.consensus_params("1")
    assert "result" in resp


@pytest.mark.asyncio
async def test_tx(noderpc):
    resp = await noderpc.tx("36F0945A22CD6921FF9F85F64080334B4887DE4021B2EA5EE7B1D182C3FFEE01")
    assert "error" in resp


@pytest.mark.asyncio
async def test_tx_search(noderpc):
    resp = await noderpc.tx_search("tx.height=1000", per_page=1)
    assert "result" in resp


@pytest.mark.asyncio
async def test_unconfirmed_txs(noderpc):
    resp = await noderpc.unconfirmed_txs()
    assert "result" in resp


@pytest.mark.asyncio
async def test_validators(noderpc):
    resp = await noderpc.validators()
    assert "result" in resp


@pytest.mark.asyncio
async def test_invalid_post_request(noderpc):
    noderpc.url = "https://binance.org/invalid"
    try:
        resp = await noderpc.post_request('INVALID')
        assert False, resp
    except BinanceChainException as e:
        assert e.response.status == 403
        assert isinstance(e.__cause__, aiohttp.ContentTypeError)


@pytest.mark.asyncio
async def test_invalid_get_request(noderpc):
    try:
        resp = await noderpc.get_request('INVALID')
        assert False, resp
    except BinanceChainException as e:
        assert e.response.status == 404
        assert isinstance(e.__cause__, aiohttp.ContentTypeError)


@pytest.mark.asyncio
async def test_del_without_close_warning():
    noderpc = NodeRPC(testnet=True)
    await noderpc.get_status()
    with pytest.warns(UserWarning):
        del(noderpc)


@pytest.mark.asyncio
async def test_ws_open_close(noderpc):
    """"Open then immediately close"""
    def on_open():
        print('opened')
        noderpc.close()

    await noderpc.start_async(on_open=on_open, on_error=on_error)
    print('closed')


@pytest.mark.asyncio
async def test_ws_subscribe(noderpc):
    """"Open then immediately close with the decorator API"""

    results = []

    def on_msg(msg):
        results.append(msg)
        noderpc.close()

    def on_open():
        noderpc.subscribe("tm.event = 'Tx'")

    await noderpc.start_async(on_open=on_open, on_error=on_error, on_msg=on_msg)
    assert results
    result_ids = [r['id'] for r in results]
    assert result_ids == ['0']
    for result in results:
        assert result['result'] == {}


@pytest.mark.asyncio
async def test_ws_unsubscribe(noderpc):
    """"Open then immediately close with the decorator API"""

    results = []

    def on_msg(msg):
        print('on_msg', msg)
        results.append(msg)
        if len(results) == 1:
            noderpc.unsubscribe("tm.event = 'Tx'")
        if len(results) == 2:
            noderpc.close()

    def on_open():
        noderpc.subscribe("tm.event = 'Tx'")

    await noderpc.start_async(on_open=on_open, on_error=on_error, on_msg=on_msg)
    assert results
    result_ids = [r['id'] for r in results]
    assert result_ids == ['0', '1']
    for result in results:
        assert result['result'] == {}


@pytest.mark.asyncio
async def test_ws_unsubscribe_all(noderpc):
    """"Open then immediately close with the decorator API"""

    results = []

    def on_msg(msg):
        print('on_msg', msg)
        results.append(msg)
        if len(results) == 1:
            noderpc.unsubscribe_all()
        if len(results) == 2:
            noderpc.close()

    def on_open():
        noderpc.subscribe("tm.event = 'Tx'")

    await noderpc.start_async(on_open=on_open, on_error=on_error, on_msg=on_msg)
    assert results
    result_ids = [r['id'] for r in results]
    assert result_ids == ['0', '1']
    for result in results:
        assert result['result'] == {}



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance DEX SDK Test Suite for Transaction builder
"""
from pprint import pprint

import pytest
import asyncio
import json
from decimal import Decimal

from binancechain import Transaction, Wallet, HTTPClient, BinanceChainException
from binancechain.enums import Side, Votes, Ordertype, Timeinforce

MNEMONIC_2 = "tennis utility midnight pattern that foot security tent punch glance still night virus loop trade velvet rent glare ramp cushion defy grass section cage"
MNEMONIC = "apart conduct congress bless remember picnic aerobic nothing dinner guilt catch brain sunny vocal advice castle horror shift reject valley evoke fork syrup code"
PAIR = "IBB-8DE_BNB"
PROPOSAL_ID = 370


@pytest.fixture
async def client():
    client = HTTPClient(testnet=True)
    yield client
    await client.close()


@pytest.fixture
async def wallet():
    wallet = Wallet.wallet_from_mnemonic(words=MNEMONIC, testnet=True)
    yield wallet


@pytest.fixture
async def wallet_two():
    wallet_two = Wallet.wallet_from_mnemonic(words=MNEMONIC_2, testnet=True)
    yield wallet_two


@pytest.mark.asyncio
async def test_new_order(wallet, client):
    address = wallet.get_address()
    transaction = await Transaction.new_order_transaction(
        address=address,
        symbol=PAIR,
        side=Side.BUY,
        timeInForce=Timeinforce.GTE,
        price=0.01,
        quantity=1,
        ordertype=Ordertype.LIMIT,
        client=client,
    )
    pubkey, signature = wallet.sign(transaction.get_sign_message())
    hex_data = transaction.update_signature(pubkey, signature)
    broadcast = await client.broadcast(hex_data)
    txid = broadcast[0]["hash"]
    assert txid
    await asyncio.sleep(2)
    tx = await client.get_transaction(txid)
    assert tx["data"], "transaction not found"


@pytest.mark.asyncio
async def test_cancel_order(wallet, client):
    address = wallet.get_address()
    open_orders = await client.get_open_orders(address)
    order = open_orders["order"][0]
    refid = order["orderId"]
    symbol = order["symbol"]
    transaction = await Transaction.cancel_order_transaction(
        address=address, symbol=symbol, refid=refid, client=client
    )
    pubkey, signature = wallet.sign(transaction.get_sign_message())
    hex_data = transaction.update_signature(pubkey, signature)
    broadcast = await client.broadcast(hex_data)
    txid = broadcast[0]["hash"]
    await asyncio.sleep(2)
    tx = await client.get_transaction(txid)
    assert tx["data"], "transaction not found"


@pytest.mark.asyncio
async def test_transfer_order(wallet, wallet_two, client):
    address_1 = wallet.get_address()
    address_2 = wallet_two.get_address()
    account_1 = await client.get_account(address_1)
    account_2 = await client.get_account(address_2)
    balances_1 = account_1["balances"]
    balances_2 = account_2["balances"]
    TOKEN = "BNB"
    for balance in balances_1:
        if balance["symbol"] == TOKEN:
            balance_1 = Decimal(balance["free"])
            break
    for balance in balances_2:
        if balance["symbol"] == TOKEN:
            balance_2 = Decimal(balance["free"])
            break
    assert balance_1 and balance_1 > 0.1, "No Token balance to test"
    transaction = await Transaction.transfer_transaction(
        from_address=address_1,
        to_address=address_2,
        symbol=TOKEN,
        amount=0.1,
        client=client,
    )
    pubkey, signature = wallet.sign(transaction.get_sign_message())
    hex_data = transaction.update_signature(pubkey, signature)
    broadcast = await client.broadcast(hex_data)
    assert broadcast, "Fail to broadcast"
    assert "hash" in broadcast[0], "No txid"
    txid = broadcast[0]["hash"]
    await asyncio.sleep(1)
    tx = await client.get_transaction(txid)
    assert tx, "No transaction on client"
    assert "data" in tx, "tx is not on chain"
    print(tx)
    # TODO check balance again


@pytest.mark.asyncio
async def test_freeze_token(wallet, client):
    address = wallet.get_address()
    transaction = await Transaction.freeze_token_transaction(
        address=address, symbol="BNB", amount=0.0001, client=client
    )
    pubkey, signature = wallet.sign(transaction.get_sign_message())
    hex_data = transaction.update_signature(pubkey, signature)
    broadcast = await client.broadcast(hex_data)
    txid = broadcast[0]["hash"]
    await asyncio.sleep(1)
    tx = await client.get_transaction(txid)


@pytest.mark.asyncio
async def test_unfreeze_token(wallet, client):
    address = wallet.get_address()
    transaction = await Transaction.unfreeze_token_transaction(
        address=address, symbol="BNB", amount=0.0001, client=client
    )
    pubkey, signature = wallet.sign(transaction.get_sign_message())
    hex_data = transaction.update_signature(pubkey, signature)
    broadcast = await client.broadcast(hex_data)
    txid = broadcast[0]["hash"]
    await asyncio.sleep(1)
    tx = await client.get_transaction(txid)


@pytest.mark.asyncio
async def test_transaction_object_new_order(wallet, client):
    address = wallet.get_address()
    transaction = Transaction(wallet=wallet, client=client)
    new_order = await transaction.create_new_order(
        symbol=PAIR,
        side=Side.BUY,
        timeInForce=Timeinforce.GTE,
        price=0.01,
        quantity=1,
        ordertype=Ordertype.LIMIT,
    )
    assert new_order, "No result of transfer found"
    assert new_order[0]["hash"], "No txid found"


@pytest.mark.asyncio
async def test_transaction_object_cancel(wallet, client):
    address = wallet.get_address()
    open_orders = await client.get_open_orders(address)
    order = open_orders["order"][0]
    refid = order["orderId"]
    symbol = order["symbol"]
    transaction = Transaction(wallet=wallet, client=client)
    cancel_order = await transaction.cancel_order(symbol=symbol, refid=refid)
    assert cancel_order, "No result of transfer found"
    assert cancel_order[0]["hash"], "No txid found"


@pytest.mark.asyncio
async def test_transaction_object_transfer(wallet, wallet_two, client):
    transaction = Transaction(wallet=wallet, client=client)
    transfer = await transaction.transfer(
        to_address=wallet_two.get_address(), symbol="BNB", amount=0.1
    )
    assert transfer, "No result of transfer found"
    assert transfer[0]["hash"], "No txid found"


@pytest.mark.asyncio
async def test_transaction_object_freeze(wallet, client):
    transaction = Transaction(wallet=wallet, client=client)
    freeze = await transaction.freeze_token(symbol="BNB", amount=0.1)
    assert freeze, "No result of transfer found"
    assert freeze[0]["hash"], "No txid found"


@pytest.mark.asyncio
async def test_transaction_object_unfreeze(wallet, client):
    transaction = Transaction(wallet=wallet, client=client)
    unfreeze = await transaction.unfreeze_token(symbol="BNB", amount=0.1)
    assert unfreeze, "No result of transfer found"
    assert unfreeze[0]["hash"], "No txid found"


# @pytest.mark.asyncio
# async def test_vote_token(wallet, client):
#     address = wallet.get_address()
#     transaction = await Transaction.vote_transaction(
#         voter=address, proposal_id=PROPOSAL_ID, option=Votes.YES, client=client
#     )
#     pubkey, signature = wallet.sign(transaction.get_sign_message())
#     hex_data = transaction.update_signature(pubkey, signature)
#     broadcast = await client.broadcast(hex_data)
#     txid = broadcast[0]["hash"]
#     await asyncio.sleep(1)
#     tx = await client.get_transaction(txid)
#     print(tx)



#!/usr/bin/env python

import setuptools
from distutils.core import setup

setup(
    name="binancechain",
    version="0.1.4",
    description="Unofficial Binance Chain SDK",
    author="Luke Macken & Kim Bui",
    author_email="",
    url="https://github.com/lmacken/binance-chain-python",
    packages=["binancechain"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: AsyncIO",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    install_requires=[
        "wheel",
        "bech32",
        "aiohttp",
        "bitcoinlib",
        "eth_keyfile",
        "secp256k1",
        "pyee",
        "varint",
        "protobuf",
        "click",
    ],
)



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT
"""
Binance Chain HTTP API

https://docs.binance.org/api-reference/dex-api/paths.html
"""
import logging
import warnings
from typing import Any, Dict, List, Optional, Tuple

import aiohttp

from .exceptions import BinanceChainException

log = logging.getLogger(__name__)

MAINNET_URL = "https://dex.binance.org"
TESTNET_URL = "https://testnet-dex.binance.org"


class HTTPClient:
    """ Binance Chain HTTP API Client """

    def __init__(self, testnet: bool = True, api_version: str = "v1", url=None):
        """
        :param testnet: Use testnet instead of mainnet
        :param api_version: The API version to use
        :param session: An optional HTTP session to use
        """
        if not url:
            url = TESTNET_URL if testnet else MAINNET_URL
        self._server = f"{url}/api/{api_version}/"
        self._session: aiohttp.ClientSession = None
        self._testnet = testnet

    def __del__(self):
        if self._session:  # pragma: nocover
            warnings.warn(f"{repr(self)}.close() was never awaited")

    async def close(self):
        """ Clean up our connections """
        if self._session:
            await self._session.close()
            self._session = None

    async def _request(self, method: str, path: str, **kwargs):
        """
        :method: `get` or `post`
        :path: the remote endpoint to call
        :kwargs: Extra arguments to pass to the request, like `params` or `data`.
        :raises: `BinanceChainException`, which has a `response` attribute.
        """
        if not self._session:
            self._session = aiohttp.ClientSession()
        try:
            resp = None
            async with getattr(self._session, method)(
                self._server + path, **kwargs
            ) as resp:
                return await resp.json()
        except Exception as e:
            log.exception(f"Request error: {method} {path} {kwargs}")
            raise BinanceChainException(resp) from e

    async def get_request(self, path: str, params: dict = None) -> Any:
        """Perform a GET request"""
        return await self._request("get", path, params=params)

    async def post_request(
        self, path: str, data: Optional[str] = None, headers: Optional[dict] = None
    ) -> Any:
        """Perform a POST request"""
        return await self._request("post", path, data=data, headers=headers)

    async def get_time(self) -> dict:
        """Get the block time.

        Gets the latest block time and the current time according to the HTTP
        service.

        Destination: Validator node.
        Rate Limit: 1 request per IP per second.
        """
        return await self.get_request("time")

    async def get_node_info(self) -> dict:
        """Get node information.

        Gets runtime information about the node.

        Destination: Validator node.
        Rate Limit: 1 request per IP per second.
        """
        return await self.get_request("node-info")

    async def get_validators(self) -> dict:
        """Get validators.

        Gets the list of validators used in consensus.

        Destination: Witness node.
        Rate Limit: 10 requests per IP per second.
        """
        return await self.get_request("validators")

    async def get_peers(self) -> List[dict]:
        """Get network peers.

        Gets the list of network peers.

        Destination: Witness node.
        Rate Limit: 1 request per IP per second.
        """
        return await self.get_request("peers")

    async def get_account(self, address: str) -> dict:
        """Get an account.

        Gets account metadata for an address.

        Destination: Witness node.
        Rate Limit: 5 requests per IP per second.

        :param address: The account address to query
        """
        return await self.get_request(f"account/{address}")

    async def get_account_sequence(self, address: str) -> dict:
        """Get an account sequence.

        Gets an account sequence for an address.

        Destination: Validator node.
        Rate Limit: 5 requests per IP per second.
        Parameters:

        :param address: The account address to query.
        """
        return await self.get_request(f"account/{address}/sequence")

    async def get_transaction(self, hash: str) -> dict:
        """Get a transaction.

        Gets transaction metadata by transaction ID. By default, transactions
        are returned in a raw format.

        Destination: Seed node.
        Rate Limit: 10 requests per IP per second.

        :param hash: The transaction hash to query
        """
        return await self.get_request(f"tx/{hash}")

    async def get_token_list(self) -> List[dict]:
        """Get tokens list.

        Gets a list of tokens that have been issued.

        Destination: Witness node.
        Rate Limit: 1 request per IP per second.
        """
        return await self.get_request("tokens")

    async def get_markets(self, limit: int = 500, offset: int = 0) -> List[dict]:
        """Get market pairs.

        Gets the list of market pairs that have been listed.

        Destination: Witness node.
        Rate Limit: 1 request per IP per second.

        :param limit: default 500; max 1000.
        :param offset: start with 0; default 0.
        """
        return await self.get_request(
            "markets", params={"limit": limit, "offset": offset}
        )

    async def get_fees(self) -> List[dict]:
        """Obtain trading fees information.

        Gets the current trading fees settings.

        Destination: Witness node.
        Rate Limit: 1 request per IP per second.
        """
        return await self.get_request("fees")

    async def get_depth(self, symbol: str, limit: int = 100) -> Dict[str, list]:
        """Get the order book.

        Gets the order book depth data for a given pair symbol.
        The given limit must be one of the allowed limits below.

        Destination: Validator node.
        Rate Limit: 10 requests per IP per second.

        :param symbol: Market pair symbol, e.g. NNB-0AD_BNB
        :param limit: The limit of results. Allowed limits: [5, 10, 20, 50, 100, 500, 1000]
        """
        return await self.get_request(
            "depth", params={"symbol": symbol, "limit": limit}
        )

    async def broadcast(self, body: str, sync: bool = None) -> List[dict]:
        """Broadcast a transaction.

        Broadcasts a signed transaction. A single transaction must be sent
        hex-encoded with a content-type of text/plain.

        Destination: Witness node.
        Rate Limit: 5 requests per IP per second.

        :param sync: Synchronous broadcast (wait for DeliverTx)?
        :param body: Hex-encoded transaction
        """
        return await self.post_request(
            "broadcast", data=body, headers={"Content-Type": "text/plain"}
        )

    async def get_klines(
        self,
        symbol: str,
        interval: str,
        limit: int = 300,
        start: int = None,
        end: int = None,
    ) -> List[Tuple[int, str, str, str, str, str, int, str, int]]:
        """Get candlestick bars.

        Gets candlestick/kline bars for a symbol. Bars are uniquely identified
        by their open time.

        If the time window is larger than limits, only the first n klines will
        return. In this case, please either shrink the window or increase the
        limit to get proper amount of klines.

        Rate Limit: 10 requests per IP per second.

        :param symbol: symbol
        :param interval: interval. Allowed value: [1m, 3m, 5m, 15m, 30m, 1h,
            2h, 4h, 6h, 8h, 12h, 1d, 3d, 1w, 1M]
        :param limit: default 300; max 1000.
        :param start: start time in Milliseconds
        :param end: end time in Milliseconds
        """
        params = {"symbol": symbol, "interval": interval, "limit": limit}
        if start is not None:
            params["startTime"] = int(start)
        if end is not None:
            params["endTime"] = int(end)
        return await self.get_request("klines", params=params)

    async def get_closed_orders(
        self,
        address: str,
        end: int = None,
        limit: int = None,
        offset: int = None,
        side: int = None,
        start: int = None,
        status: str = None,
        symbol: str = None,
        total: int = None,
    ) -> List[dict]:
        """Get closed orders.

        Gets closed (filled and cancelled) orders for a given address.

        Query Window: Default query window is latest 7 days; The maximum
        start - end query window is 3 months.
        Rate Limit: 5 requests per IP per second.

        :param address: the owner address
        :param end: end time in Milliseconds
        :param limit: default 500; max 1000.
        :param offset: start with 0; default 0.
        :param side: order side. 1 for buy and 2 for sell.
        :param start: start time in Milliseconds
        :param status: order status list. Allowed value: [Ack, PartialFill,
            IocNoFill, FullyFill, Canceled, Expired, FailedBlocking,
            FailedMatching]
        :param symbol: symbol
        :param total: total number required, 0 for not required and 1 for
            required; default not required, return total=-1 in response
        """
        params = {
            "address": address,
        }
        if end is not None:
            params['end'] = int(end)
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params['offset'] = offset
        if side is not None:
            params['side'] = side
        if start is not None:
            params['start'] = int(start)
        if status is not None:
            params['status'] = status
        if symbol is not None:
            params['symbol'] = symbol
        if total is not None:
            params['total'] = total
        return await self.get_request("orders/closed", params=params)

    async def get_open_orders(
        self,
        address: str,
        limit: int = None,
        offset: int = None,
        symbol: str = None,
        total: int = None,
    ):
        """Get open orders.

        Gets open orders for a given address.

        Rate Limit: 5 requests per IP per second.

        :param address: the owner address
        :param limit: default 500; max 1000.
        :param offset: start with 0; default 0.
        :param symbol: symbol
        :param total: total number required, 0 for not required and 1 for
            required; default not required, return total=-1 in response
        """
        params = {"address": address}
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if symbol is not None:
            params["symbol"] = symbol
        if total is not None:
            params["total"] = total
        return await self.get_request("orders/open", params=params)

    async def get_order(self, id: str) -> dict:
        """Get an order.

        Gets metadata for an individual order by its ID.

        Rate Limit: 5 requests per IP per second.

        :param id: order id
        """
        return await self.get_request(f"orders/{id}")

    async def get_ticker(self, symbol: str = None) -> List[dict]:
        """Get a market ticker.

        Gets 24 hour price change statistics for a market pair symbol.

        Rate Limit: 5 requests per IP per second.

        :param symbol: symbol
        """
        params = {}
        if symbol:
            params["symbol"] = symbol
        return await self.get_request("ticker/24hr", params=params)

    async def get_trades(
        self,
        address: str = None,
        buyerOrderId: str = None,
        height: int = None,
        limit: int = None,
        offset: int = None,
        quoteAsset: str = None,
        sellerOrderId: str = None,
        side: int = None,
        start: int = None,
        end: int = None,
        total: int = None,
        symbol: str = None,
    ) -> List[dict]:
        """Get market trades.

        Gets a list of historical trades.

        Query Window: Default query window is latest 7 days; The maximum
        start - end query window is 3 months.

        Rate Limit: 5 requests per IP per second.

        :param address: the buyer/seller address
        :param buyerOrderId: buyer order id
        :param end: end time in Milliseconds
        :param height: block height
        :param limit: default 500; max 1000.
        :param start: start with 0; default 0.
        :param quoteAsset: quote asset
        :param sellerOrderId: seller order id
        :param side: order side. 1 for buy and 2 for sell.
        :param start: start time in Milliseconds
        :param symbol: symbol
        :param total: total number required, 0 for not required and 1 for
            required; default not required, return total=-1 in response
        """
        params: Dict[Any, Any] = {}
        if address is not None:
            params["address"] = address
        if buyerOrderId is not None:
            params["buyerOrderId"] = buyerOrderId
        if height is not None:
            params["height"] = height
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if sellerOrderId is not None:
            params["sellerOrderId"] = sellerOrderId
        if side is not None:
            params["side"] = side
        if start is not None:
            params["start"] = start
        if end is not None:
            params["end"] = end
        if symbol is not None:
            params["symbol"] = symbol
        if total is not None:
            params["total"] = total
        return await self.get_request("trades", params=params)

    async def get_block_exchange_fee(
        self,
        address: str = None,
        end: int = None,
        limit: int = None,
        offset: int = None,
        start: int = None,
        total: int = None,
    ) -> List[dict]:
        """Trading fee of the address grouped by block

        Get historical trading fees of the address, including fees of
        trade/canceled order/expired order. Transfer and other transaction fees
        are not included. Order by block height DESC.

        Query Window: Default query window is latest 7 days; The maximum
        start - end query window is 3 months.

        Rate Limit: 5 requests per IP per second.

        :param address: the buyer/seller address
        :param buyerOrderId: buyer order id
        :param end: end time in Milliseconds
        :param limit: default 500; max 1000.
        :param offset: start with 0; default 0.
        :param start: start with 0; default 0.
        :param total: total number required, 0 for not required and 1 for
            required; default not required, return total=-1 in response
        """
        params: Dict[Any, Any] = {}
        if address is not None:
            params["address"] = address
        if end is not None:
            params["end"] = end
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if start is not None:
            params["start"] = start
        if total is not None:
            params["total"] = total
        return await self.get_request("block-exchange-fee", params=params)

    async def get_transactions(
        self,
        address: str,
        height: int = None,
        end: int = None,
        limit: int = None,
        offset: int = None,
        side: str = None,
        start: int = None,
        tx_asset: str = None,
        tx_type: str = None,
    ) -> List[dict]:
        """Get transactions.

        Gets a list of transactions. Multisend transaction is not available in
        this API.

        Query Window: Default query window is latest 24 hours; The maximum
        start - end query window is 3 months.

        Rate Limit: 60 requests per IP per minute.

        :param address: address
        :param height: block height
        :param end: end time in milliseconds
        :param limit: limit
        :param offset: offset
        :param side: transaction side. Allowed value: [ RECEIVE, SEND]
        :param start: start time in milliseconds
        :param tx_asset: txAsset
        :param tx_type: transaction type. Allowed value: [
            NEW_ORDER,ISSUE_TOKEN,BURN_TOKEN,LIST_TOKEN,CANCEL_ORDER,FREEZE_TOKEN,
            UN_FREEZE_TOKEN,TRANSFER,PROPOSAL,VOTE,MINT,DEPOSIT]
        """
        params: Dict[Any, Any] = {"address": address}
        if height is not None:
            params["blockHeight"] = height
        if start is not None:
            params["startTime"] = start
        if end is not None:
            params["endTime"] = end
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if side is not None:
            params["side"] = side
        if tx_asset is not None:
            params["txAsset"] = tx_asset
        if tx_type is not None:
            params["txType"] = tx_type
        return await self.get_request("transactions", params=params)



# Copyright 2019, Luke Macken, Kim Bui, and the binance-chain-python contributors
# SPDX-License-Identifier: MIT

from enum import Enum


class Votes(Enum):
    YES = 1
    NO = 2
    ABSTAIN = 3
    NOWITHVETO = 4


class Ordertype(Enum):
    LIMIT = 2


class Side(Enum):
    BUY = 1
    SELL = 2


class Timeinforce(Enum):
    GTE = 1
    IOC = 3
